from os.path import dirname
from os.path import join

import numpy as np
import scipy.fftpack
import scipy.io
import scipy.misc
import scipy.ndimage
import scipy.stats

import cv2
from .save_stats import aggd_features,paired_product,compute_image_mscn_transform

np.int = int
np.float = np.float64
np.bool = np.bool_

def _extract_subband_feats(mscncoefs):
    # alpha_m,  = extract_ggd_features(mscncoefs)
    alpha_m, N, bl, br, lsq, rsq = aggd_features(mscncoefs.copy())
    pps1, pps2, pps3, pps4 = paired_product(mscncoefs)
    alpha1, N1, bl1, br1, lsq1, rsq1 = aggd_features(pps1)
    alpha2, N2, bl2, br2, lsq2, rsq2 = aggd_features(pps2)
    alpha3, N3, bl3, br3, lsq3, rsq3 = aggd_features(pps3)
    alpha4, N4, bl4, br4, lsq4, rsq4 = aggd_features(pps4)
    return np.array([alpha_m, (bl+br)/2.0]), np.array([
            alpha1, N1, bl1, br1,  # (V)
            alpha2, N2, bl2, br2,  # (H)
            alpha3, N3, bl3, br3,  # (D1)
            alpha4, N4, bl4, br4,  # (D2)
    ])

def extract_on_patches(img, blocksizerow, blocksizecol):
    h, w = img.shape
    blocksizerow = np.int(blocksizerow)
    blocksizecol = np.int(blocksizecol)
    patches = []
    for j in range(0, np.int(h-blocksizerow+1), np.int(blocksizerow)):
        for i in range(0, np.int(w-blocksizecol+1), np.int(blocksizecol)):
            patch = img[j:j+blocksizerow, i:i+blocksizecol]
            patches.append(patch)

    patches = np.array(patches)
    
    patch_features = []
    for p in patches:
        mscn_features, pp_features = _extract_subband_feats(p)
        patch_features.append(np.hstack((mscn_features, pp_features)))
    patch_features = np.array(patch_features)

    return patch_features

def computequality(img, blocksizerow, blocksizecol, mu_prisparam, cov_prisparam):
    # img = img[:, :, 0]
    h, w = img.shape

    if (h < blocksizerow) or (w < blocksizecol):
        print("Input frame is too small")
        exit(0)

    # ensure that the patch divides evenly into img
    hoffset = (h % blocksizerow)
    woffset = (w % blocksizecol)

    if hoffset > 0: 
        img = img[:-hoffset, :]
    if woffset > 0:
        img = img[:, :-woffset]

    img = img.astype(np.float32)

    scale_percent = 50 # percent of original size
    width = int(img.shape[1] * scale_percent / 100)
    height = int(img.shape[0] * scale_percent / 100) 

    img2 = cv2.resize(img, (height,width), interpolation=cv2.INTER_CUBIC)

    mscn1, var, mu = compute_image_mscn_transform(img, extend_mode='nearest')
    mscn1 = mscn1.astype(np.float32)

    mscn2, _, _ = compute_image_mscn_transform(img2, extend_mode='nearest')
    mscn2 = mscn2.astype(np.float32)

    feats_lvl1 = extract_on_patches(mscn1, blocksizerow, blocksizecol)
    feats_lvl2 = extract_on_patches(mscn2, blocksizerow/2, blocksizecol/2)

    # stack the scale features
    feats = np.hstack((feats_lvl1, feats_lvl2))# feats_lvl3))

    mu_distparam = np.mean(feats, axis=0)
    cov_distparam = np.cov(feats.T)

    invcov_param = np.linalg.pinv((cov_prisparam + cov_distparam)/2)

    xd = mu_prisparam - mu_distparam 
    quality = np.sqrt(np.dot(np.dot(xd, invcov_param), xd.T))[0][0]

    return np.hstack((mu_distparam, [quality]))


def compute_niqe_features(frames):
    blocksizerow = 96
    blocksizecol = 96

    M, N = frames.shape
    assert ((M >= blocksizerow*2) & (N >= blocksizecol*2)), "Video too small for NIQE extraction"

    module_path = dirname(__file__)
    params = scipy.io.loadmat(join(module_path, 'frames_modelparameters.mat'))
    mu_prisparam = params['mu_prisparam']
    cov_prisparam = params['cov_prisparam']

    # niqe_features = np.zeros((frames.shape[0]-10, 37))
    # idx = 0
    # for i in range(5, frames.shape[0]-5):
    #   niqe_features[idx] = computequality(frames[i], blocksizerow, blocksizecol, mu_prisparam, cov_prisparam)
    #   idx += 1

    niqe_features = computequality(frames, blocksizerow, blocksizecol, mu_prisparam, cov_prisparam)
    return niqe_features




"""Tests for davis_jf module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_davis_jf_basics():
    from ayase.modules.davis_jf import DAVISJFModule
    _test_module_basics(DAVISJFModule, "davis_jf")

def test_davis_jf_no_reference(image_sample):
    from ayase.modules.davis_jf import DAVISJFModule
    m = DAVISJFModule()
    result = m.process(image_sample)
    assert result is image_sample



"""KVD (Kernel Video Distance) module.

KVD is an alternative to FVD that uses kernel methods (Maximum Mean Discrepancy)
instead of Gaussian assumptions. Better for non-Gaussian feature distributions.
Lower KVD = better video generation quality.

This is a dataset-level metric that compares two distributions of videos.
"""

import logging
from typing import Optional, List

import numpy as np

from ayase.models import Sample
from ayase.base_modules import BatchMetricModule

logger = logging.getLogger(__name__)


class KVDModule(BatchMetricModule):
    name = "kvd"
    description = "Kernel Video Distance using Maximum Mean Discrepancy (batch metric)"
    default_config = {
        "feature_extractor": "i3d",  # Same as FVD
        "kernel": "rbf",  # RBF kernel for MMD
        "bandwidth": 1.0,  # Kernel bandwidth
        "device": "auto",
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.feature_extractor_type = self.config.get("feature_extractor", "i3d")
        self.kernel_type = self.config.get("kernel", "rbf")
        self.bandwidth = self.config.get("bandwidth", 1.0)
        self.device_config = self.config.get("device", "auto")
        self.device = None
        self._ml_available = False
        self._feature_model = None

    def setup(self) -> None:
        try:
            import torch

            # Set device
            if self.device_config == "auto":
                self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            else:
                self.device = torch.device(self.device_config)

            # Load feature extractor (reuse FVD's I3D)
            from ayase.modules.fvd import FVDModule

            fvd_module = FVDModule(self.config)
            fvd_module.setup()
            self._feature_model = fvd_module._r3d_model
            self._ml_available = fvd_module._ml_available

            if self._ml_available:
                logger.info(f"KVD module initialized on {self.device}")

        except Exception as e:
            logger.warning(f"Failed to setup KVD: {e}")

    def extract_features(self, sample: Sample) -> Optional[np.ndarray]:
        """Extract features using same method as FVD."""
        if not sample.is_video or self._feature_model is None:
            return None

        # Reuse FVD's feature extraction
        from ayase.modules.fvd import FVDModule

        fvd_temp = FVDModule(self.config)
        fvd_temp._r3d_model = self._feature_model
        fvd_temp._ml_available = True
        return fvd_temp.extract_features(sample)

    def _rbf_kernel(self, X: np.ndarray, Y: np.ndarray) -> np.ndarray:
        """Compute RBF (Gaussian) kernel matrix.

        Args:
            X: Feature matrix (N, D)
            Y: Feature matrix (M, D)

        Returns:
            Kernel matrix (N, M)
        """
        # Compute pairwise squared distances
        XX = np.sum(X ** 2, axis=1)[:, np.newaxis]
        YY = np.sum(Y ** 2, axis=1)[np.newaxis, :]
        XY = X @ Y.T
        dists = XX + YY - 2 * XY

        # RBF kernel
        K = np.exp(-dists / (2 * self.bandwidth ** 2))
        return K

    def _compute_mmd(self, X: np.ndarray, Y: np.ndarray) -> float:
        """Compute Maximum Mean Discrepancy.

        Args:
            X: Features from distribution 1
            Y: Features from distribution 2

        Returns:
            MMD score (lower = more similar distributions)
        """
        # Kernel matrices
        K_XX = self._rbf_kernel(X, X)
        K_YY = self._rbf_kernel(Y, Y)
        K_XY = self._rbf_kernel(X, Y)

        m = X.shape[0]
        n = Y.shape[0]

        # MMD^2 = E[K(X,X)] + E[K(Y,Y)] - 2*E[K(X,Y)]
        mmd_sq = (K_XX.sum() - np.trace(K_XX)) / (m * (m - 1))
        mmd_sq += (K_YY.sum() - np.trace(K_YY)) / (n * (n - 1))
        mmd_sq -= 2 * K_XY.mean()

        return float(max(mmd_sq, 0.0))  # MMD^2 can be slightly negative due to numerical errors

    def compute_distribution_metric(
        self, features: List[np.ndarray], reference_features: Optional[List[np.ndarray]] = None
    ) -> float:
        """Compute KVD using Maximum Mean Discrepancy.

        Args:
            features: List of feature vectors from generated/test videos
            reference_features: Optional list of features from real/reference videos

        Returns:
            KVD score (lower is better)
        """
        try:
            # Convert to numpy array
            features_array = np.stack(features, axis=0)

            if reference_features is not None and len(reference_features) > 0:
                ref_array = np.stack(reference_features, axis=0)
            else:
                # Split features in half
                mid = len(features_array) // 2
                ref_array = features_array[:mid]
                features_array = features_array[mid:]

            # Compute MMD
            mmd = self._compute_mmd(features_array, ref_array)

            # Scale to similar range as FVD (multiply by large constant)
            kvd = mmd * 1000.0

            return float(kvd)

        except Exception as e:
            logger.error(f"Failed to compute KVD: {e}")
            return float('inf')

    def on_dispose(self) -> None:
        """Compute KVD after all samples processed."""
        if len(self._feature_cache) < 2:
            logger.info(f"KVD: Not enough samples ({len(self._feature_cache)}) for metric computation")
            self._feature_cache = []
            self._reference_cache = []
            return

        try:
            kvd_score = self.compute_distribution_metric(
                self._feature_cache,
                self._reference_cache if self._reference_cache else None
            )

            logger.info(
                f"KVD computed: {kvd_score:.2f} "
                f"(generated: {len(self._feature_cache)}, "
                f"reference: {len(self._reference_cache)})"
            )

            # Store in pipeline stats if available
            if hasattr(self, "pipeline") and self.pipeline:
                if hasattr(self.pipeline, "add_dataset_metric"):
                    self.pipeline.add_dataset_metric("kvd", kvd_score)

        except Exception as e:
            logger.error(f"Failed to compute KVD: {e}")

        finally:
            self._feature_cache = []
            self._reference_cache = []



# Copyright (c) Facebook, Inc. and its affiliates.
# Modified by Jeffrey Ouyang-Zhang

from typing import List

import torch
import torch.nn as nn

from util.box_ops import (
    box_cxcywh_to_xyxy,
    box_iou,
    box_xyxy_to_cxcywh,
    generalized_box_iou,
)


# from https://github.com/facebookresearch/detectron2/blob/cbbc1ce26473cb2a5cc8f58e8ada9ae14cb41052/detectron2/layers/wrappers.py#L100
def nonzero_tuple(x):
    """
    A 'as_tuple=True' version of torch.nonzero to support torchscript.
    because of https://github.com/pytorch/pytorch/issues/38718
    """
    if torch.jit.is_scripting():
        if x.dim() == 0:
            return x.unsqueeze(0).nonzero().unbind(1)
        return x.nonzero().unbind(1)
    else:
        return x.nonzero(as_tuple=True)


# from https://github.com/facebookresearch/detectron2/blob/9921a2caa585d4fa66c4b534b6fab6e74d89b582/detectron2/modeling/matcher.py#L9
class Matcher(object):
    """
    This class assigns to each predicted "element" (e.g., a box) a ground-truth
    element. Each predicted element will have exactly zero or one matches; each
    ground-truth element may be matched to zero or more predicted elements.

    The matching is determined by the MxN match_quality_matrix, that characterizes
    how well each (ground-truth, prediction)-pair match each other. For example,
    if the elements are boxes, this matrix may contain box intersection-over-union
    overlap values.

    The matcher returns (a) a vector of length N containing the index of the
    ground-truth element m in [0, M) that matches to prediction n in [0, N).
    (b) a vector of length N containing the labels for each prediction.
    """

    def __init__(
        self,
        thresholds: List[float],
        labels: List[int],
        allow_low_quality_matches: bool = False,
    ):
        """
        Args:
            thresholds (list): a list of thresholds used to stratify predictions
                into levels.
            labels (list): a list of values to label predictions belonging at
                each level. A label can be one of {-1, 0, 1} signifying
                {ignore, negative class, positive class}, respectively.
            allow_low_quality_matches (bool): if True, produce additional matches
                for predictions with maximum match quality lower than high_threshold.
                See set_low_quality_matches_ for more details.

            For example,
                thresholds = [0.3, 0.5]
                labels = [0, -1, 1]
                All predictions with iou < 0.3 will be marked with 0 and
                thus will be considered as false positives while training.
                All predictions with 0.3 <= iou < 0.5 will be marked with -1 and
                thus will be ignored.
                All predictions with 0.5 <= iou will be marked with 1 and
                thus will be considered as true positives.
        """
        # Add -inf and +inf to first and last position in thresholds
        thresholds = thresholds[:]
        assert thresholds[0] > 0
        thresholds.insert(0, -float("inf"))
        thresholds.append(float("inf"))
        # Currently torchscript does not support all + generator
        assert all(
            [low <= high for (low, high) in zip(thresholds[:-1], thresholds[1:])]
        ), thresholds
        assert all([l in [-1, 0, 1] for l in labels])
        assert len(labels) == len(thresholds) - 1
        self.thresholds = thresholds
        self.labels = labels
        self.allow_low_quality_matches = allow_low_quality_matches

    def __call__(self, match_quality_matrix):
        """
        Args:
            match_quality_matrix (Tensor[float]): an MxN tensor, containing the
                pairwise quality between M ground-truth elements and N predicted
                elements. All elements must be >= 0 (due to the us of `torch.nonzero`
                for selecting indices in :meth:`set_low_quality_matches_`).

        Returns:
            matches (Tensor[int64]): a vector of length N, where matches[i] is a matched
                ground-truth index in [0, M)
            match_labels (Tensor[int8]): a vector of length N, where pred_labels[i] indicates
                whether a prediction is a true or false positive or ignored
        """
        assert match_quality_matrix.dim() == 2
        if match_quality_matrix.numel() == 0:
            default_matches = match_quality_matrix.new_full(
                (match_quality_matrix.size(1),), 0, dtype=torch.int64
            )
            # When no gt boxes exist, we define IOU = 0 and therefore set labels
            # to `self.labels[0]`, which usually defaults to background class 0
            # To choose to ignore instead, can make labels=[-1,0,-1,1] + set appropriate thresholds
            default_match_labels = match_quality_matrix.new_full(
                (match_quality_matrix.size(1),), self.labels[0], dtype=torch.int8
            )
            return default_matches, default_match_labels

        assert torch.all(match_quality_matrix >= 0)

        # match_quality_matrix is M (gt) x N (predicted)
        # Max over gt elements (dim 0) to find best gt candidate for each prediction
        matched_vals, matches = match_quality_matrix.max(dim=0)

        match_labels = matches.new_full(matches.size(), 1, dtype=torch.int8)

        for l, low, high in zip(self.labels, self.thresholds[:-1], self.thresholds[1:]):
            low_high = (matched_vals >= low) & (matched_vals < high)
            match_labels[low_high] = l

        if self.allow_low_quality_matches:
            self.set_low_quality_matches_(match_labels, match_quality_matrix)

        return matches, match_labels

    def set_low_quality_matches_(self, match_labels, match_quality_matrix):
        """
        Produce additional matches for predictions that have only low-quality matches.
        Specifically, for each ground-truth G find the set of predictions that have
        maximum overlap with it (including ties); for each prediction in that set, if
        it is unmatched, then match it to the ground-truth G.

        This function implements the RPN assignment case (i) in Sec. 3.1.2 of
        :paper:`Faster R-CNN`.
        """
        # For each gt, find the prediction with which it has highest quality
        highest_quality_foreach_gt, _ = match_quality_matrix.max(dim=1)
        # Find the highest quality match available, even if it is low, including ties.
        # Note that the matches qualities must be positive due to the use of
        # `torch.nonzero`.
        _, pred_inds_with_highest_quality = nonzero_tuple(
            match_quality_matrix == highest_quality_foreach_gt[:, None]
        )
        # If an anchor was labeled positive only due to a low-quality match
        # with gt_A, but it has larger overlap with gt_B, it's matched index will still be gt_B.
        # This follows the implementation in Detectron, and is found to have no significant impact.
        match_labels[pred_inds_with_highest_quality] = 1


# from https://github.com/facebookresearch/detectron2/blob/cbbc1ce26473cb2a5cc8f58e8ada9ae14cb41052/detectron2/modeling/sampling.py#L9
def subsample_labels(
    labels: torch.Tensor, num_samples: int, positive_fraction: float, bg_label: int
):
    """
    Return `num_samples` (or fewer, if not enough found)
    random samples from `labels` which is a mixture of positives & negatives.
    It will try to return as many positives as possible without
    exceeding `positive_fraction * num_samples`, and then try to
    fill the remaining slots with negatives.

    Args:
        labels (Tensor): (N, ) label vector with values:
            * -1: ignore
            * bg_label: background ("negative") class
            * otherwise: one or more foreground ("positive") classes
        num_samples (int): The total number of labels with value >= 0 to return.
            Values that are not sampled will be filled with -1 (ignore).
        positive_fraction (float): The number of subsampled labels with values > 0
            is `min(num_positives, int(positive_fraction * num_samples))`. The number
            of negatives sampled is `min(num_negatives, num_samples - num_positives_sampled)`.
            In order words, if there are not enough positives, the sample is filled with
            negatives. If there are also not enough negatives, then as many elements are
            sampled as is possible.
        bg_label (int): label index of background ("negative") class.

    Returns:
        pos_idx, neg_idx (Tensor):
            1D vector of indices. The total length of both is `num_samples` or fewer.
    """
    positive = nonzero_tuple((labels != -1) & (labels != bg_label))[0]
    negative = nonzero_tuple(labels == bg_label)[0]

    num_pos = int(num_samples * positive_fraction)
    # protect against not enough positive examples
    num_pos = min(positive.numel(), num_pos)
    num_neg = num_samples - num_pos
    # protect against not enough negative examples
    num_neg = min(negative.numel(), num_neg)

    # randomly select positive and negative examples
    perm1 = torch.randperm(positive.numel(), device=positive.device)[:num_pos]
    perm2 = torch.randperm(negative.numel(), device=negative.device)[:num_neg]

    pos_idx = positive[perm1]
    neg_idx = negative[perm2]
    return pos_idx, neg_idx


def sample_topk_per_gt(pr_inds, gt_inds, iou, k):
    if len(gt_inds) == 0:
        return pr_inds, gt_inds
    # find topk matches for each gt
    gt_inds2, counts = gt_inds.unique(return_counts=True)
    scores, pr_inds2 = iou[gt_inds2].topk(k, dim=1)
    gt_inds2 = gt_inds2[:, None].repeat(1, k)

    # filter to as many matches that gt has
    pr_inds3 = torch.cat([pr[:c] for c, pr in zip(counts, pr_inds2)])
    gt_inds3 = torch.cat([gt[:c] for c, gt in zip(counts, gt_inds2)])
    return pr_inds3, gt_inds3


# modified from https://github.com/facebookresearch/detectron2/blob/cbbc1ce26473cb2a5cc8f58e8ada9ae14cb41052/detectron2/modeling/roi_heads/roi_heads.py#L123
class Stage2Assigner(nn.Module):
    def __init__(self, num_queries, max_k=4):
        super().__init__()
        self.positive_fraction = 0.25
        self.bg_label = 400  # number > 91 to filter out later
        self.batch_size_per_image = num_queries
        self.proposal_matcher = Matcher(
            thresholds=[0.6], labels=[0, 1], allow_low_quality_matches=True
        )
        self.k = max_k

    def _sample_proposals(
        self,
        matched_idxs: torch.Tensor,
        matched_labels: torch.Tensor,
        gt_classes: torch.Tensor,
    ):
        """
        Based on the matching between N proposals and M groundtruth,
        sample the proposals and set their classification labels.

        Args:
            matched_idxs (Tensor): a vector of length N, each is the best-matched
                gt index in [0, M) for each proposal.
            matched_labels (Tensor): a vector of length N, the matcher's label
                (one of cfg.MODEL.ROI_HEADS.IOU_LABELS) for each proposal.
            gt_classes (Tensor): a vector of length M.

        Returns:
            Tensor: a vector of indices of sampled proposals. Each is in [0, N).
            Tensor: a vector of the same length, the classification label for
                each sampled proposal. Each sample is labeled as either a category in
                [0, num_classes) or the background (num_classes).
        """
        has_gt = gt_classes.numel() > 0
        # Get the corresponding GT for each proposal
        if has_gt:
            gt_classes = gt_classes[matched_idxs]
            # Label unmatched proposals (0 label from matcher) as background (label=num_classes)
            gt_classes[matched_labels == 0] = self.bg_label
            # Label ignore proposals (-1 label)
            gt_classes[matched_labels == -1] = -1
        else:
            gt_classes = torch.zeros_like(matched_idxs) + self.bg_label

        sampled_fg_idxs, sampled_bg_idxs = subsample_labels(
            gt_classes, self.batch_size_per_image, self.positive_fraction, self.bg_label
        )

        sampled_idxs = torch.cat([sampled_fg_idxs, sampled_bg_idxs], dim=0)
        return sampled_idxs, gt_classes[sampled_idxs]

    def forward(self, outputs, targets, return_cost_matrix=False):
        # COCO categories are from 1 to 90. They set num_classes=91 and apply sigmoid.

        bs = len(targets)
        indices = []
        ious = []
        for b in range(bs):
            iou, _ = box_iou(
                box_cxcywh_to_xyxy(targets[b]["boxes"]),
                box_cxcywh_to_xyxy(outputs["init_reference"][b].detach()),
            )
            matched_idxs, matched_labels = self.proposal_matcher(
                iou
            )  # proposal_id -> highest_iou_gt_id, proposal_id -> [1 if iou > 0.6, 0 ow]
            sampled_idxs, sampled_gt_classes = (
                self._sample_proposals(  # list of sampled proposal_ids, sampled_id -> [0, num_classes)+[bg_label]
                    matched_idxs, matched_labels, targets[b]["labels"]
                )
            )
            pos_pr_inds = sampled_idxs[sampled_gt_classes != self.bg_label]
            pos_gt_inds = matched_idxs[pos_pr_inds]
            pos_pr_inds, pos_gt_inds = self.postprocess_indices(
                pos_pr_inds, pos_gt_inds, iou
            )
            indices.append((pos_pr_inds, pos_gt_inds))
            ious.append(iou)
        if return_cost_matrix:
            return indices, ious
        return indices

    def postprocess_indices(self, pr_inds, gt_inds, iou):
        return sample_topk_per_gt(pr_inds, gt_inds, iou, self.k)


# modified from https://github.com/facebookresearch/detectron2/blob/cbbc1ce26473cb2a5cc8f58e8ada9ae14cb41052/detectron2/modeling/proposal_generator/rpn.py#L181
class Stage1Assigner(nn.Module):
    def __init__(self, t_low=0.3, t_high=0.7, max_k=4):
        super().__init__()
        self.positive_fraction = 0.5
        self.batch_size_per_image = 256
        self.k = max_k
        self.t_low = t_low
        self.t_high = t_high
        self.anchor_matcher = Matcher(
            thresholds=[t_low, t_high],
            labels=[0, -1, 1],
            allow_low_quality_matches=True,
        )

    def _subsample_labels(self, label):
        """
        Randomly sample a subset of positive and negative examples, and overwrite
        the label vector to the ignore value (-1) for all elements that are not
        included in the sample.

        Args:
            labels (Tensor): a vector of -1, 0, 1. Will be modified in-place and returned.
        """
        pos_idx, neg_idx = subsample_labels(
            label, self.batch_size_per_image, self.positive_fraction, 0
        )
        # Fill with the ignore label (-1), then set positive and negative labels
        label.fill_(-1)
        label.scatter_(0, pos_idx, 1)
        label.scatter_(0, neg_idx, 0)
        return label

    def forward(self, outputs, targets):
        bs = len(targets)
        indices = []
        for b in range(bs):
            anchors = outputs["anchors"][b]
            if len(targets[b]["boxes"]) == 0:
                indices.append(
                    (
                        torch.tensor([], dtype=torch.long, device=anchors.device),
                        torch.tensor([], dtype=torch.long, device=anchors.device),
                    )
                )
                continue
            iou, _ = box_iou(
                box_cxcywh_to_xyxy(targets[b]["boxes"]),
                box_cxcywh_to_xyxy(anchors),
            )
            matched_idxs, matched_labels = self.anchor_matcher(
                iou
            )  # proposal_id -> highest_iou_gt_id, proposal_id -> [1 if iou > 0.7, 0 if iou < 0.3, -1 ow]
            matched_labels = self._subsample_labels(matched_labels)

            all_pr_inds = torch.arange(len(anchors)).to(anchors.device)

            pos_pr_inds = all_pr_inds[matched_labels == 1]
            pos_gt_inds = matched_idxs[pos_pr_inds]
            pos_ious = iou[pos_gt_inds, pos_pr_inds]
            pos_pr_inds, pos_gt_inds = self.postprocess_indices(
                pos_pr_inds, pos_gt_inds, iou
            )
            pos_pr_inds, pos_gt_inds = pos_pr_inds.to(anchors.device), pos_gt_inds.to(
                anchors.device
            )
            indices.append((pos_pr_inds, pos_gt_inds))
        return indices

    def postprocess_indices(self, pr_inds, gt_inds, iou):
        return sample_topk_per_gt(pr_inds, gt_inds, iou, self.k)



"""
Code from https://github.com/mlfoundations/wise-ft/blob/master/src/datasets/imagenetv2.py
Thanks to the authors of wise-ft
"""

import pathlib
import shutil
import tarfile

import requests
from PIL import Image
from torch.utils.data import DataLoader, Dataset
from torchvision.datasets import ImageFolder
from tqdm import tqdm

URLS = {
    "matched-frequency": "https://imagenetv2public.s3-us-west-2.amazonaws.com/imagenetv2-matched-frequency.tar.gz",
    "threshold-0.7": "https://imagenetv2public.s3-us-west-2.amazonaws.com/imagenetv2-threshold0.7.tar.gz",
    "top-images": "https://imagenetv2public.s3-us-west-2.amazonaws.com/imagenetv2-top-images.tar.gz",
    "val": "https://imagenetv2public.s3-us-west-2.amazonaws.com/imagenet_validation.tar.gz",
}

FNAMES = {
    "matched-frequency": "imagenetv2-matched-frequency-format-val",
    "threshold-0.7": "imagenetv2-threshold0.7-format-val",
    "top-images": "imagenetv2-top-images-format-val",
    "val": "imagenet_validation",
}


V2_DATASET_SIZE = 10000
VAL_DATASET_SIZE = 50000


class ImageNetValDataset(Dataset):
    def __init__(self, transform=None, location="."):
        self.dataset_root = pathlib.Path(f"{location}/imagenet_validation/")
        self.tar_root = pathlib.Path(f"{location}/imagenet_validation.tar.gz")
        self.fnames = list(self.dataset_root.glob("**/*.JPEG"))
        self.transform = transform
        if not self.dataset_root.exists() or len(self.fnames) != VAL_DATASET_SIZE:
            if not self.tar_root.exists():
                print(f"Dataset imagenet-val not found on disk, downloading....")
                response = requests.get(URLS["val"], stream=True)
                total_size_in_bytes = int(response.headers.get("content-length", 0))
                block_size = 1024  # 1 Kibibyte
                progress_bar = tqdm(
                    total=total_size_in_bytes, unit="iB", unit_scale=True
                )
                with open(self.tar_root, "wb") as f:
                    for data in response.iter_content(block_size):
                        progress_bar.update(len(data))
                        f.write(data)
                progress_bar.close()
                if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
                    assert False, f"Downloading from {URLS[variant]} failed"
            print("Extracting....")
            tarfile.open(self.tar_root).extractall(f"{location}")
            shutil.move(f"{location}/{FNAMES['val']}", self.dataset_root)

        self.dataset = ImageFolder(self.dataset_root)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, i):
        img, label = self.dataset[i]
        if self.transform is not None:
            img = self.transform(img)
        return img, label


class ImageNetV2Dataset(Dataset):
    def __init__(self, variant="matched-frequency", transform=None, location="."):
        self.dataset_root = pathlib.Path(f"{location}/ImageNetV2-{variant}/")
        self.tar_root = pathlib.Path(f"{location}/ImageNetV2-{variant}.tar.gz")
        self.fnames = list(self.dataset_root.glob("**/*.jpeg"))
        self.transform = transform
        assert variant in URLS, f"unknown V2 Variant: {variant}"
        if not self.dataset_root.exists() or len(self.fnames) != V2_DATASET_SIZE:
            if not self.tar_root.exists():
                print(f"Dataset {variant} not found on disk, downloading....")
                response = requests.get(URLS[variant], stream=True)
                total_size_in_bytes = int(response.headers.get("content-length", 0))
                block_size = 1024  # 1 Kibibyte
                progress_bar = tqdm(
                    total=total_size_in_bytes, unit="iB", unit_scale=True
                )
                with open(self.tar_root, "wb") as f:
                    for data in response.iter_content(block_size):
                        progress_bar.update(len(data))
                        f.write(data)
                progress_bar.close()
                if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
                    assert False, f"Downloading from {URLS[variant]} failed"
            print("Extracting....")
            tarfile.open(self.tar_root).extractall(f"{location}")
            shutil.move(f"{location}/{FNAMES[variant]}", self.dataset_root)
            self.fnames = list(self.dataset_root.glob("**/*.jpeg"))

    def __len__(self):
        return len(self.fnames)

    def __getitem__(self, i):
        img, label = Image.open(self.fnames[i]), int(self.fnames[i].parent.name)
        if self.transform is not None:
            img = self.transform(img)
        return img, label



# Description: This file contains the code for serializing the dataset.
# From https://github.com/ppwwyyxx/RAM-multiprocess-dataloader/blob/795868a37446d61412b9a58dbb1b7c76e75d39c4/serialize.py

# Copyright (c) Facebook, Inc. and its affiliates.
"""
List serialization code adopted from
https://github.com/facebookresearch/detectron2/blob/main/detectron2/data/common.py
"""

import multiprocessing as mp

from typing import List, Any, Optional

import pickle
import numpy as np
import torch
import torch.distributed as dist

import functools
import os

from datetime import timedelta


def get_world_size() -> int:
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size()


def get_rank() -> int:
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    return dist.get_rank()


def get_local_rank() -> int:
    if not dist.is_available():
        return 0
    if not dist.is_initialized():
        return 0
    
    # this is not guaranteed to be set
    if 'RANK' in os.environ and 'WORLD_SIZE' in os.environ:
        return int(os.environ['LOCAL_RANK'])
    elif 'SLURM_PROCID' in os.environ:
        return int(os.environ['SLURM_LOCALID'])
    else:
        raise RuntimeError("Unable to get local rank")


def get_local_size() -> int:
    return torch.cuda.device_count()


@functools.lru_cache()
def _get_global_gloo_group():
    """
    Return a process group based on gloo backend, containing all the ranks
    The result is cached.
    """
    if dist.get_backend() == "nccl":
        return dist.new_group(backend="gloo", timeout=timedelta(minutes=60))
    else:
        return dist.group.WORLD


def all_gather(data, group=None):
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors).

    Args:
        data: any picklable object
        group: a torch process group. By default, will use a group which
            contains all ranks on gloo backend.

    Returns:
        list[data]: list of data gathered from each rank
    """
    if get_world_size() == 1:
        return [data]
    if group is None:
        group = (
            _get_global_gloo_group()
        )  # use CPU group by default, to reduce GPU RAM usage.
    world_size = dist.get_world_size(group)
    if world_size == 1:
        return [data]

    output = [None for _ in range(world_size)]
    dist.all_gather_object(output, data, group=group)
    return output


class NumpySerializedList:
    def __init__(self, lst: list):
        def _serialize(data):
            buffer = pickle.dumps(data, protocol=-1)
            return np.frombuffer(buffer, dtype=np.uint8)

        print(
            "Serializing {} elements to byte tensors and concatenating them all ...".format(
                len(lst)
            )
        )
        self._lst = [_serialize(x) for x in lst]
        self._addr = np.asarray([len(x) for x in self._lst], dtype=np.int64)
        self._addr = np.cumsum(self._addr)
        self._lst = np.concatenate(self._lst)
        print("Serialized dataset takes {:.2f} MiB".format(len(self._lst) / 1024**2))

    def __len__(self):
        return len(self._addr)

    def __getitem__(self, idx):
        start_addr = 0 if idx == 0 else self._addr[idx - 1].item()
        end_addr = self._addr[idx].item()
        bytes = memoryview(self._lst[start_addr:end_addr])
        return pickle.loads(bytes)


class TorchSerializedList(NumpySerializedList):
    def __init__(self, lst: list):
        super().__init__(lst)
        self._addr = torch.from_numpy(self._addr)
        self._lst = torch.from_numpy(self._lst)

    def __getitem__(self, idx):
        start_addr = 0 if idx == 0 else self._addr[idx - 1].item()
        end_addr = self._addr[idx].item()
        bytes = memoryview(self._lst[start_addr:end_addr].numpy())
        return pickle.loads(bytes)


def local_scatter(array: Optional[List[Any]]):
    """
    Scatter an array from local leader to all local workers.
    The i-th local worker gets array[i].

    Args:
        array: Array with same size of #local workers.
    """
    if get_local_size() <= 1:
        # Just one worker. Do nothing.
        return array[0]
    if get_local_rank() == 0:
        assert len(array) == get_local_size()
        all_gather(array)
    else:
        all_data = all_gather(None)
        array = all_data[get_rank() - get_local_rank()]
    return array[get_local_rank()]


# NOTE: https://github.com/facebookresearch/mobile-vision/pull/120
# has another implementation that does not use tensors.
class TorchShmSerializedList(TorchSerializedList):
    def __init__(self, lst: list):
        if get_local_rank() == 0:
            super().__init__(lst)

        if get_local_rank() == 0:
            # Move data to shared memory, obtain a handle to send to each local worker.
            # This is cheap because a tensor will only be moved to shared memory once.
            handles = [None] + [
                bytes(mp.reduction.ForkingPickler.dumps((self._addr, self._lst)))
                for _ in range(get_local_size() - 1)
            ]
        else:
            handles = None
        # Each worker receives the handle from local leader.
        handle = local_scatter(handles)

        if get_local_rank() > 0:
            # Materialize the tensor from shared memory.
            self._addr, self._lst = mp.reduction.ForkingPickler.loads(handle)
            print(
                f"Worker {get_rank()} obtains a dataset of length="
                f"{len(self)} from its local leader."
            )


# From https://github.com/ppwwyyxx/RAM-multiprocess-dataloader/issues/5#issuecomment-1510676170
def local_broadcast_process_authkey():
    if int(os.environ['LOCAL_WORLD_SIZE']) == 1:
        return 
    local_rank = int(os.environ['LOCAL_RANK'])
    authkey = bytes(mp.current_process().authkey)
    all_keys = all_gather(authkey)
    local_leader_key = all_keys[get_rank() - local_rank]
    if authkey != local_leader_key:
        print("Process authkey is different from the key of local leader. This might happen when "
              "workers are launched independently.")
        print("Overwriting local authkey ...")
        mp.current_process().authkey = local_leader_key



"""Semantic Segmentation Consistency module.

Measures temporal stability of semantic segmentation across video
frames:

  semantic_consistency — 0-1 (higher = more consistent)

Algorithm:
  1. Produce per-frame semantic segmentation (SegFormer or simple
     colour-based clustering as fallback).
  2. Compute frame-to-frame segment overlap (IoU per class).
  3. Average IoU across frames and classes.

Videos only — images are skipped.
"""

import logging
from pathlib import Path
from typing import List, Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class SemanticSegmentationConsistencyModule(PipelineModule):
    name = "semantic_segmentation_consistency"
    description = "Temporal stability of semantic segmentation"
    default_config = {
        "backend": "auto",  # "segformer", "kmeans", or "auto"
        "device": "auto",
        "subsample": 3,
        "max_frames": 150,
        "num_clusters": 8,  # For K-means fallback
        "warning_threshold": 0.6,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.backend = self.config.get("backend", "auto")
        self.device_config = self.config.get("device", "auto")
        self.subsample = self.config.get("subsample", 3)
        self.max_frames = self.config.get("max_frames", 150)
        self.num_clusters = self.config.get("num_clusters", 8)
        self.warning_threshold = self.config.get("warning_threshold", 0.6)

        self.device = None
        self._segformer_model = None
        self._segformer_processor = None
        self._use_segformer = False
        self._ml_available = False

    def setup(self) -> None:
        if self.backend in ("segformer", "auto"):
            try:
                import torch
                from transformers import SegformerForSemanticSegmentation, SegformerImageProcessor

                if self.device_config == "auto":
                    self.device = "cuda" if torch.cuda.is_available() else "cpu"
                else:
                    self.device = self.device_config

                self._segformer_processor = SegformerImageProcessor.from_pretrained(
                    "nvidia/segformer-b0-finetuned-ade-512-512"
                )
                self._segformer_model = SegformerForSemanticSegmentation.from_pretrained(
                    "nvidia/segformer-b0-finetuned-ade-512-512"
                ).to(self.device).eval()

                self._use_segformer = True
                self._ml_available = True
                logger.info(f"Semantic consistency: SegFormer initialised on {self.device}")
                return

            except ImportError:
                if self.backend == "segformer":
                    logger.warning("transformers not installed for SegFormer")
                    return
                logger.info("SegFormer unavailable, falling back to K-means")
            except Exception as e:
                logger.warning(f"SegFormer init failed: {e}")
                if self.backend == "segformer":
                    return

        # K-means fallback (always available)
        self._ml_available = True
        logger.info("Semantic consistency: K-means colour clustering backend")

    # ------------------------------------------------------------------
    # Segmentation
    # ------------------------------------------------------------------

    def _segment_segformer(self, frame_bgr: np.ndarray) -> Optional[np.ndarray]:
        """Return integer label map (H, W) using SegFormer."""
        import torch

        rgb = cv2.cvtColor(frame_bgr, cv2.COLOR_BGR2RGB)
        from PIL import Image
        pil_img = Image.fromarray(rgb)

        inputs = self._segformer_processor(images=pil_img, return_tensors="pt")
        inputs = {k: v.to(self.device) for k, v in inputs.items()}

        with torch.no_grad():
            outputs = self._segformer_model(**inputs)
            logits = outputs.logits  # (1, C, H', W')

        # Upsample to original size
        h, w = frame_bgr.shape[:2]
        upsampled = torch.nn.functional.interpolate(
            logits, size=(h, w), mode="bilinear", align_corners=False
        )
        labels = upsampled.argmax(dim=1).squeeze(0).cpu().numpy().astype(np.int32)
        return labels

    def _segment_kmeans(self, frame_bgr: np.ndarray) -> np.ndarray:
        """Return integer label map (H, W) using K-means colour clustering."""
        h, w = frame_bgr.shape[:2]

        # Downsample for speed
        small = cv2.resize(frame_bgr, (w // 2, h // 2))
        lab = cv2.cvtColor(small, cv2.COLOR_BGR2LAB).astype(np.float32)
        pixels = lab.reshape(-1, 3)

        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 20, 1.0)
        _, labels_flat, _ = cv2.kmeans(
            pixels, self.num_clusters, None, criteria, 3, cv2.KMEANS_PP_CENTERS
        )

        labels = labels_flat.reshape(h // 2, w // 2).astype(np.int32)
        # Upsample back
        labels = cv2.resize(labels, (w, h), interpolation=cv2.INTER_NEAREST)
        return labels

    def _segment(self, frame_bgr: np.ndarray) -> Optional[np.ndarray]:
        try:
            if self._use_segformer:
                return self._segment_segformer(frame_bgr)
            return self._segment_kmeans(frame_bgr)
        except Exception as e:
            logger.debug(f"Segmentation failed: {e}")
            return None

    # ------------------------------------------------------------------
    # IoU computation
    # ------------------------------------------------------------------

    @staticmethod
    def _mean_iou(labels_a: np.ndarray, labels_b: np.ndarray) -> float:
        """Compute mean IoU between two label maps."""
        classes = set(np.unique(labels_a)) | set(np.unique(labels_b))
        if not classes:
            return 1.0

        ious = []
        for c in classes:
            mask_a = labels_a == c
            mask_b = labels_b == c
            intersection = np.logical_and(mask_a, mask_b).sum()
            union = np.logical_or(mask_a, mask_b).sum()
            if union > 0:
                ious.append(float(intersection / union))

        return float(np.mean(ious)) if ious else 0.0

    # ------------------------------------------------------------------
    # Process
    # ------------------------------------------------------------------

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available or not sample.is_video:
            return sample

        try:
            return self._process_video(sample)
        except Exception as e:
            logger.error(f"Semantic consistency failed for {sample.path}: {e}")
            return sample

    def _process_video(self, sample: Sample) -> Sample:
        cap = cv2.VideoCapture(str(sample.path))
        if not cap.isOpened():
            return sample

        label_maps: List[np.ndarray] = []
        idx = 0

        while idx < self.max_frames:
            ret, frame = cap.read()
            if not ret:
                break
            if idx % self.subsample == 0:
                labels = self._segment(frame)
                if labels is not None:
                    label_maps.append(labels)
            idx += 1

        cap.release()

        if len(label_maps) < 2:
            return sample

        # Compute frame-to-frame mean IoU
        ious = []
        for i in range(1, len(label_maps)):
            iou = self._mean_iou(label_maps[i - 1], label_maps[i])
            ious.append(iou)

        if not ious:
            return sample

        consistency = float(np.mean(ious))

        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        sample.quality_metrics.semantic_consistency = consistency

        if consistency < self.warning_threshold:
            sample.validation_issues.append(
                ValidationIssue(
                    severity=ValidationSeverity.WARNING,
                    message=f"Low semantic consistency: {consistency:.3f}",
                    details={
                        "semantic_consistency": consistency,
                        "min_iou": float(min(ious)),
                    },
                    recommendation=(
                        "Semantic segmentation is unstable between frames. "
                        "Object boundaries may be flickering."
                    ),
                )
            )

        logger.debug(
            f"Semantic consistency for {sample.path.name}: "
            f"{consistency:.3f} (min={min(ious):.3f})"
        )

        return sample



#!/usr/bin/env python
#
# File Name : ptbtokenizer.py
#
# Description : Do the PTB Tokenization and remove punctuations.
#
# Creation Date : 29-12-2014
# Last Modified : Thu Mar 19 09:53:35 2015
# Authors : Hao Fang <hfang@uw.edu> and Tsung-Yi Lin <tl483@cornell.edu>
import os
import subprocess
import tempfile

# path to the stanford corenlp jar
STANFORD_CORENLP_3_4_1_JAR = os.path.dirname(os.path.abspath(__file__)) + '/stanford-corenlp-3.4.1.jar'

# punctuations to be removed from the sentences
PUNCTUATIONS = ["''", "'", "``", "`", "-LRB-", "-RRB-", "-LCB-", "-RCB-", \
        ".", "?", "!", ",", ":", "-", "--", "...", ";"]

class PTBTokenizer:
    """Python wrapper of Stanford PTBTokenizer"""

    def tokenize(self, captions_for_image):
        cmd = [os.getenv("JAVA_HOME"), '-cp', STANFORD_CORENLP_3_4_1_JAR, \
                'edu.stanford.nlp.process.PTBTokenizer', \
                '-preserveLines', '-lowerCase']

        # ======================================================
        # prepare data for PTB Tokenizer
        # ======================================================
        final_tokenized_captions_for_image = {}
        image_id = [k for k, v in captions_for_image.items() for _ in range(len(v))]
        sentences = '\n'.join([c['caption'].replace('\n', ' ') for k, v in captions_for_image.items() for c in v])

        # ======================================================
        # save sentences to temporary file
        # ======================================================
        path_to_jar_dirname=os.path.dirname(os.path.abspath(__file__))
        tmp_file = tempfile.NamedTemporaryFile(delete=False, dir=path_to_jar_dirname)
        tmp_file.write(sentences.encode())
        tmp_file.close()

        # ======================================================
        # tokenize sentence
        # ======================================================
        cmd.append(os.path.basename(tmp_file.name))
        p_tokenizer = subprocess.Popen(cmd, cwd=path_to_jar_dirname, \
                stdout=subprocess.PIPE)
        token_lines = p_tokenizer.communicate(input=sentences.rstrip())[0]
        token_lines = token_lines.decode()
        lines = token_lines.split('\n')
        # remove temp file
        os.remove(tmp_file.name)

        # ======================================================
        # create dictionary for tokenized captions
        # ======================================================
        for k, line in zip(image_id, lines):
            if not k in final_tokenized_captions_for_image:
                final_tokenized_captions_for_image[k] = []
            tokenized_caption = ' '.join([w for w in line.rstrip().split(' ') \
                    if w not in PUNCTUATIONS])
            final_tokenized_captions_for_image[k].append(tokenized_caption)

        return final_tokenized_captions_for_image



"""MSSWD — Multi-Scale Sliced Wasserstein Distance.

Dataset-level distribution metric via pyiqa. Compares feature
distributions at multiple scales using sliced Wasserstein distance.
Lower score indicates more similar distributions.

msswd_score — LOWER = better (closer distributions)
"""

import logging
from typing import Optional, List

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import BatchMetricModule

logger = logging.getLogger(__name__)


class MSSWDModule(BatchMetricModule):
    name = "msswd"
    description = "MSSWD multi-scale sliced Wasserstein distance via pyiqa (batch, lower=better)"
    default_config = {
        "num_scales": 3,
        "num_projections": 128,
        "feature_dim": 64,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.num_scales = self.config.get("num_scales", 3)
        self.num_projections = self.config.get("num_projections", 128)
        self.feature_dim = self.config.get("feature_dim", 64)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try pyiqa MSSWD
        try:
            import pyiqa
            self._model = pyiqa.create_metric("msswd", device="cpu")
            self._backend = "pyiqa"
            logger.info("MSSWD (pyiqa) initialised")
            return
        except (ImportError, Exception):
            pass

        # Tier 2: Heuristic fallback
        self._backend = "heuristic"
        logger.info("MSSWD (heuristic) initialised — install pyiqa for full model")

    def extract_features(self, sample: Sample) -> Optional[np.ndarray]:
        """Extract multi-scale spatial features from a sample."""
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return None
                mid = total // 2
                cap.set(cv2.CAP_PROP_POS_FRAMES, mid)
                ret, frame = cap.read()
                if not ret:
                    return None
            finally:
                cap.release()
        else:
            frame = cv2.imread(str(sample.path))
            if frame is None:
                return None

        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)

        # Multi-scale feature extraction
        features = []
        current = gray.copy()
        for scale in range(self.num_scales):
            h, w = current.shape
            if h < 8 or w < 8:
                break

            # Histogram features
            hist = np.histogram(current, bins=32, range=(0, 256))[0]
            hist = hist.astype(np.float64) / (hist.sum() + 1e-7)
            features.extend(hist.tolist())

            # Statistical moments
            features.append(np.mean(current))
            features.append(np.std(current))

            # Gradient features
            gx = cv2.Sobel(current, cv2.CV_64F, 1, 0, ksize=3)
            gy = cv2.Sobel(current, cv2.CV_64F, 0, 1, ksize=3)
            grad_mag = np.sqrt(gx ** 2 + gy ** 2)
            features.append(np.mean(grad_mag))
            features.append(np.std(grad_mag))

            # Downsample for next scale
            if current.shape[0] > 16 and current.shape[1] > 16:
                current = cv2.pyrDown(current)

        # Pad or truncate to fixed dimension
        features = np.array(features, dtype=np.float64)
        if len(features) < self.feature_dim:
            features = np.pad(features, (0, self.feature_dim - len(features)))
        else:
            features = features[: self.feature_dim]

        return features

    def compute_distribution_metric(
        self, features: List, reference_features: Optional[List] = None
    ) -> float:
        """Compute multi-scale sliced Wasserstein distance between distributions."""
        feat_matrix = np.array(features)

        if reference_features is not None and len(reference_features) >= 2:
            ref_matrix = np.array(reference_features)
        else:
            # Self-comparison: split into two halves
            n = len(feat_matrix)
            if n < 4:
                return 0.0
            mid = n // 2
            ref_matrix = feat_matrix[mid:]
            feat_matrix = feat_matrix[:mid]

        # Sliced Wasserstein Distance
        np.random.seed(42)
        projections = np.random.randn(self.num_projections, feat_matrix.shape[1])
        projections = projections / (
            np.linalg.norm(projections, axis=1, keepdims=True) + 1e-7
        )

        swd_total = 0.0
        for proj in projections:
            proj_feat = feat_matrix @ proj
            proj_ref = ref_matrix @ proj
            proj_feat_sorted = np.sort(proj_feat)
            proj_ref_sorted = np.sort(proj_ref)

            # Interpolate to same length
            n_points = max(len(proj_feat_sorted), len(proj_ref_sorted))
            x_feat = np.linspace(0, 1, len(proj_feat_sorted))
            x_ref = np.linspace(0, 1, len(proj_ref_sorted))
            x_common = np.linspace(0, 1, n_points)

            interp_feat = np.interp(x_common, x_feat, proj_feat_sorted)
            interp_ref = np.interp(x_common, x_ref, proj_ref_sorted)

            swd_total += np.mean((interp_feat - interp_ref) ** 2)

        swd = swd_total / self.num_projections
        return float(np.clip(swd, 0.0, 1e6))



"""Tests for KID (Kernel Inception Distance) module."""

import numpy as np
import pytest

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics, DatasetStats


def test_kid_basics():
    from ayase.modules.kid import KIDModule

    _test_module_basics(KIDModule, "kid")


def test_kid_skip_without_ml(image_sample):
    from ayase.modules.kid import KIDModule

    m = KIDModule()
    # Don't call setup — _ml_available stays False
    result = m.process(image_sample)
    assert result is image_sample


def test_kid_feature_extraction(image_sample):
    from ayase.modules.kid import KIDModule

    m = KIDModule()
    m.setup()
    if not m._ml_available:
        pytest.skip("ML dependencies not available")
    feat = m.extract_features(image_sample)
    # For native backend, should return a numpy array
    if m._backend == "native":
        assert isinstance(feat, np.ndarray)
        assert feat.shape == (2048,)
    else:
        # For cleanfid/torch-fidelity, returns path string
        assert feat is not None


def test_kid_compute_same_distribution():
    """KID between two identical feature sets should be approximately 0."""
    from ayase.modules.kid import KIDModule

    m = KIDModule({"subset_size": 10, "num_subsets": 10})
    # Synthetic features
    np.random.seed(42)
    features = np.random.randn(50, 2048).astype(np.float32)

    kid_mean, kid_std = m._compute_kid_mmd(features, features)
    # KID of same distribution should be close to 0
    assert abs(kid_mean) < 0.5, f"KID of same distribution too large: {kid_mean}"


def test_kid_compute_different_distributions():
    """KID between different distributions should be > 0."""
    from ayase.modules.kid import KIDModule

    m = KIDModule({"subset_size": 10, "num_subsets": 10})
    np.random.seed(42)
    features_a = np.random.randn(50, 2048).astype(np.float32)
    features_b = np.random.randn(50, 2048).astype(np.float32) + 5.0  # Shifted

    kid_mean, _ = m._compute_kid_mmd(features_a, features_b)
    assert kid_mean > 0, f"KID should be > 0 for different distributions, got {kid_mean}"


def test_kid_dataset_stats_field():
    """Verify kid and kid_std fields exist in DatasetStats."""
    stats = DatasetStats(
        total_samples=10,
        valid_samples=10,
        invalid_samples=0,
        total_size=1000,
    )
    assert hasattr(stats, "kid")
    assert hasattr(stats, "kid_std")
    stats.kid = 0.05
    stats.kid_std = 0.01
    assert stats.kid == 0.05
    assert stats.kid_std == 0.01



"""Unified-VQA — Unified Video Quality Assessment (FR+NR Multi-task).

2025 — combines spatial sharpness, temporal consistency, and
reference similarity (when available) into a unified quality score.
Operates in NR mode when no reference is provided, FR mode otherwise.

Stores result in dover_score as proxy (shared NR quality field).
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class UnifiedVQAModule(PipelineModule):
    name = "unified_vqa"
    description = "Unified-VQA FR+NR multi-task quality assessment (2025)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try native unified_vqa
        try:
            import unified_vqa
            self._model = unified_vqa
            self._backend = "native"
            logger.info("Unified-VQA (native) initialised")
            return
        except ImportError:
            pass

        # Tier 2: Heuristic fallback
        self._backend = "heuristic"
        logger.info("Unified-VQA (heuristic) initialised — install unified_vqa for full model")

    def process(self, sample: Sample) -> Sample:
        try:
            if self._backend == "native":
                score = float(self._model.predict(str(sample.path)))
            else:
                score = self._process_heuristic(sample)

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                # Store in dover_score as proxy (shared NR quality field)
                # Only set if not already set by DOVER module
                if sample.quality_metrics.dover_score is None:
                    sample.quality_metrics.dover_score = score

        except Exception as e:
            logger.warning(f"Unified-VQA failed for {sample.path}: {e}")

        return sample

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: spatial sharpness + temporal consistency + reference similarity."""
        frames = []
        reference_path = getattr(sample, "reference_path", None)

        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return None
                indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        frames.append(frame)
            finally:
                cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)

        if not frames:
            return None

        # Component 1: Spatial sharpness
        sharpness_scores = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            lap_var = cv2.Laplacian(gray, cv2.CV_64F).var()
            sharpness_scores.append(min(lap_var / 800.0, 1.0))
        spatial_sharpness = float(np.mean(sharpness_scores))

        # Component 2: Spatial quality (contrast + noise)
        spatial_quality_scores = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            contrast = min(gray.std() / 65.0, 1.0)
            # Noise estimation via high-frequency energy
            sobel = cv2.Sobel(gray, cv2.CV_64F, 1, 1)
            noise_est = sobel.var()
            cleanliness = 1.0 / (1.0 + noise_est * 0.0001)
            spatial_quality_scores.append(0.5 * contrast + 0.5 * cleanliness)
        spatial_quality = float(np.mean(spatial_quality_scores))

        # Component 3: Temporal consistency
        if len(frames) > 1:
            grays = [
                cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(np.float64) for f in frames
            ]
            diffs = []
            for i in range(len(grays) - 1):
                diffs.append(np.mean(np.abs(grays[i + 1] - grays[i])))
            diffs = np.array(diffs)
            # Low variance in diffs = smooth temporal transitions
            temporal_consistency = 1.0 / (1.0 + np.var(diffs) * 0.01)
        else:
            temporal_consistency = 1.0

        # Component 4: Reference similarity (FR mode if reference available)
        ref_similarity = None
        if reference_path is not None:
            try:
                from pathlib import Path
                ref_path = Path(reference_path)
                if ref_path.exists():
                    ref_img = cv2.imread(str(ref_path))
                    if ref_img is not None:
                        ref_gray = cv2.cvtColor(ref_img, cv2.COLOR_BGR2GRAY).astype(np.float64)
                        dist_gray = cv2.cvtColor(frames[0], cv2.COLOR_BGR2GRAY).astype(np.float64)

                        if dist_gray.shape != ref_gray.shape:
                            dist_gray = cv2.resize(
                                dist_gray, (ref_gray.shape[1], ref_gray.shape[0])
                            )

                        # SSIM-like comparison
                        c1 = (0.01 * 255) ** 2
                        c2 = (0.03 * 255) ** 2
                        mu_d = cv2.GaussianBlur(dist_gray, (11, 11), 1.5)
                        mu_r = cv2.GaussianBlur(ref_gray, (11, 11), 1.5)
                        sigma_d2 = cv2.GaussianBlur(dist_gray ** 2, (11, 11), 1.5) - mu_d ** 2
                        sigma_r2 = cv2.GaussianBlur(ref_gray ** 2, (11, 11), 1.5) - mu_r ** 2
                        sigma_dr = (
                            cv2.GaussianBlur(dist_gray * ref_gray, (11, 11), 1.5)
                            - mu_d * mu_r
                        )
                        ssim_map = ((2 * mu_d * mu_r + c1) * (2 * sigma_dr + c2)) / (
                            (mu_d ** 2 + mu_r ** 2 + c1) * (sigma_d2 + sigma_r2 + c2)
                        )
                        ref_similarity = float(np.mean(ssim_map))
            except Exception:
                pass

        # Combine components
        if ref_similarity is not None:
            # FR mode: include reference similarity
            score = (
                0.25 * spatial_sharpness
                + 0.20 * spatial_quality
                + 0.20 * temporal_consistency
                + 0.35 * ref_similarity
            )
        else:
            # NR mode: rely on spatial and temporal features
            score = (
                0.35 * spatial_sharpness
                + 0.30 * spatial_quality
                + 0.35 * temporal_consistency
            )

        return float(np.clip(score, 0.0, 1.0))



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
Various positional encodings for the transformer.
"""
import math
import torch
from torch import nn

from util.misc import NestedTensor


class PositionEmbeddingSine(nn.Module):
    """
    This is a more standard version of the position embedding, very similar to the one
    used by the Attention is all you need paper, generalized to work on images.
    """
    def __init__(self, num_pos_feats=64, temperature=10000, normalize=False, scale=None):
        super().__init__()
        self.num_pos_feats = num_pos_feats
        self.temperature = temperature
        self.normalize = normalize
        if scale is not None and normalize is False:
            raise ValueError("normalize should be True if scale is passed")
        if scale is None:
            scale = 2 * math.pi
        self.scale = scale

    def forward(self, tensor_list: NestedTensor):
        x = tensor_list.tensors
        mask = tensor_list.mask
        assert mask is not None
        not_mask = ~mask
        y_embed = not_mask.cumsum(1, dtype=torch.float32)
        x_embed = not_mask.cumsum(2, dtype=torch.float32)
        if self.normalize:
            eps = 1e-6
            y_embed = (y_embed - 0.5) / (y_embed[:, -1:, :] + eps) * self.scale
            x_embed = (x_embed - 0.5) / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(self.num_pos_feats, dtype=torch.float32, device=x.device)
        dim_t = self.temperature ** (2 * (dim_t // 2) / self.num_pos_feats)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack((pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos_y = torch.stack((pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos = torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        return pos


class PositionEmbeddingLearned(nn.Module):
    """
    Absolute pos embedding, learned.
    """
    def __init__(self, num_pos_feats=256):
        super().__init__()
        self.row_embed = nn.Embedding(50, num_pos_feats)
        self.col_embed = nn.Embedding(50, num_pos_feats)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.uniform_(self.row_embed.weight)
        nn.init.uniform_(self.col_embed.weight)

    def forward(self, tensor_list: NestedTensor):
        x = tensor_list.tensors
        h, w = x.shape[-2:]
        i = torch.arange(w, device=x.device)
        j = torch.arange(h, device=x.device)
        x_emb = self.col_embed(i)
        y_emb = self.row_embed(j)
        pos = torch.cat([
            x_emb.unsqueeze(0).repeat(h, 1, 1),
            y_emb.unsqueeze(1).repeat(1, w, 1),
        ], dim=-1).permute(2, 0, 1).unsqueeze(0).repeat(x.shape[0], 1, 1, 1)
        return pos


def build_position_encoding(args):
    N_steps = args.hidden_dim // 2
    if args.position_embedding in ('v2', 'sine'):
        # TODO find a better way of exposing other arguments
        position_embedding = PositionEmbeddingSine(N_steps, normalize=True)
    elif args.position_embedding in ('v3', 'learned'):
        position_embedding = PositionEmbeddingLearned(N_steps)
    else:
        raise ValueError(f"not supported {args.position_embedding}")

    return position_embedding



# ------------------------------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------------------------------
# Modified from https://github.com/chengdazhi/Deformable-Convolution-V2-PyTorch/tree/pytorch_1.0.0
# ------------------------------------------------------------------------------------------------

from __future__ import absolute_import, division, print_function

import MultiScaleDeformableAttention as MSDA

import torch
import torch.nn.functional as F
from torch.autograd import Function
from torch.autograd.function import once_differentiable


class MSDeformAttnFunction(Function):
    @staticmethod
    def forward(
        ctx,
        value,
        value_spatial_shapes,
        value_level_start_index,
        sampling_locations,
        attention_weights,
        im2col_step,
    ):
        ctx.im2col_step = im2col_step
        output = MSDA.ms_deform_attn_forward(
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
            ctx.im2col_step,
        )
        ctx.save_for_backward(
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
        )
        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        (
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
        ) = ctx.saved_tensors
        grad_value, grad_sampling_loc, grad_attn_weight = MSDA.ms_deform_attn_backward(
            value,
            value_spatial_shapes,
            value_level_start_index,
            sampling_locations,
            attention_weights,
            grad_output,
            ctx.im2col_step,
        )

        return grad_value, None, None, grad_sampling_loc, grad_attn_weight, None


def ms_deform_attn_core_pytorch(
    value, value_spatial_shapes, sampling_locations, attention_weights
):
    # for debug and test only,
    # need to use cuda version instead
    N_, S_, M_, D_ = value.shape
    _, Lq_, M_, L_, P_, _ = sampling_locations.shape
    value_list = value.split([H_ * W_ for H_, W_ in value_spatial_shapes], dim=1)
    sampling_grids = 2 * sampling_locations - 1
    sampling_value_list = []
    for lid_, (H_, W_) in enumerate(value_spatial_shapes):
        # N_, H_*W_, M_, D_ -> N_, H_*W_, M_*D_ -> N_, M_*D_, H_*W_ -> N_*M_, D_, H_, W_
        value_l_ = (
            value_list[lid_].flatten(2).transpose(1, 2).reshape(N_ * M_, D_, H_, W_)
        )
        # N_, Lq_, M_, P_, 2 -> N_, M_, Lq_, P_, 2 -> N_*M_, Lq_, P_, 2
        sampling_grid_l_ = sampling_grids[:, :, :, lid_].transpose(1, 2).flatten(0, 1)
        # N_*M_, D_, Lq_, P_
        sampling_value_l_ = F.grid_sample(
            value_l_,
            sampling_grid_l_,
            mode="bilinear",
            padding_mode="zeros",
            align_corners=False,
        )
        sampling_value_list.append(sampling_value_l_)
    # (N_, Lq_, M_, L_, P_) -> (N_, M_, Lq_, L_, P_) -> (N_, M_, 1, Lq_, L_*P_)
    attention_weights = attention_weights.transpose(1, 2).reshape(
        N_ * M_, 1, Lq_, L_ * P_
    )
    output = (
        (torch.stack(sampling_value_list, dim=-2).flatten(-2) * attention_weights)
        .sum(-1)
        .view(N_, M_ * D_, Lq_)
    )
    return output.transpose(1, 2).contiguous()



import logging
import torch
from torch.utils.data import ConcatDataset, DataLoader
from .resample_concat_dataset import ResampleConcatDataset
from torchvision import transforms
from torchvision.transforms import InterpolationMode
import copy

from .sampler import StatefulDistributedSampler
from .dataloader import MetaLoader, MetaLoader_rs # NOTE keep it
from .ret_dataset import (ImgTxtRetTrainDataset,
                                 VidTxtRetTrainDataset,
                                 ImgTxtRetEvalDataset,
                                 VidTxtRetEvalDataset,
                                 VidTxtRetMCEvalDataset,
                                 VidTxtRetMCNewEvalDataset)
# from dataset.ret_dataset import (ImgTxtRetTrainDataset,
#                                  VidTxtRetTrainDataset,
#                                  ImgTxtRetEvalDataset,
#                                  VidTxtRetEvalDataset,
#                                  AudioVidTxtRetTrainDataset,
#                                  AudioVidTxtRetEvalDataset,
#                                  VidTxtRetMCEvalDataset,
#                                  AudioTxtRetTrainDataset,
#                                  AudioTxtRetEvalDataset)

from .qa_dataset import ImageQADataset, VideoQADataset
from .pt_dataset import (ImgTxtPtTrainDataset,
                                VidTxtPtTrainDataset,)
# from dataset.pt_dataset import (ImgTxtPtTrainDataset,
#                                 VidTxtPtTrainDataset,
#                                 AudioVidTxtPtTrainDataset,
#                                 AudioTxtPtTrainDataset)

logger = logging.getLogger(__name__)

def get_media_type(dataset_config):
    return dataset_config['media_type']

def get_dataset_cls(dataset_type, media_type, data_cfg):
    if dataset_type == "pt_train":
        if media_type == "image":
            dataset_cls = ImgTxtPtTrainDataset
        elif media_type == "video":
            dataset_cls = VidTxtPtTrainDataset
        # elif media_type == "audio_video":
        #     dataset_cls = AudioVidTxtPtTrainDataset
        # elif media_type == "audio":
        #     dataset_cls = AudioTxtPtTrainDataset
        else:
            raise NotImplementedError(f"dataset_type={dataset_type}, media_type={media_type}")
    elif dataset_type == "ret_train":
        if media_type == "image":
            dataset_cls = ImgTxtRetTrainDataset
        elif media_type == "video":
            dataset_cls = VidTxtRetTrainDataset
        # elif media_type == 'audio':
        #     dataset_cls = AudioTxtRetTrainDataset
        # elif media_type == "audio_video":
        #     dataset_cls = AudioVidTxtRetTrainDataset
        else:
            raise NotImplementedError(f"dataset_type={dataset_type}, media_type={media_type}")
    elif dataset_type == "ret_eval":
        if media_type == "image":
            dataset_cls = ImgTxtRetEvalDataset
        elif media_type == "video":
            dataset_cls = VidTxtRetEvalDataset
        # elif media_type == "audio":
        #     dataset_cls = AudioTxtRetEvalDataset
        # elif media_type == "audio_video":
        #     dataset_cls = AudioVidTxtRetEvalDataset
        else:
            raise NotImplementedError(f"dataset_type={dataset_type}, media_type={media_type}")
    elif dataset_type in ["qa_train", 'qa_eval']:
        if media_type == "image":
            dataset_cls = ImageQADataset
        elif media_type == "video":
            dataset_cls = VideoQADataset
        else:
            raise NotImplementedError(f"dataset_type={dataset_type}, media_type={media_type}")
    else:
        raise NotImplementedError(f"dataset_type={dataset_type}, media_type={media_type}")
    
    print(f"\033[31m dataset_type: {dataset_type} media_type: {media_type} dataset_cls: {dataset_cls}\033[0m")
    logger.info(f"dataset_type: {dataset_type} media_type: {media_type} dataset_cls: {dataset_cls}")

    return dataset_cls

def get_train_transform(config, train_file):
    vision_enc_name = config.model.vision_encoder.name
    if "internvideo" in vision_enc_name or "vit" in vision_enc_name or "umt" in vision_enc_name:
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)
    elif "clip" in vision_enc_name:
        mean = (0.48145466, 0.4578275, 0.40821073)
        std = (0.26862954, 0.26130258, 0.27577711)
    else:
        raise NotImplementedError(vision_enc_name)

    normalize = transforms.Normalize(mean, std)

    # loaded images and videos are torch.Tensor of torch.uint8 format,
    # ordered as (T, 1 or 3, H, W) where T=1 for image
    type_transform = transforms.Lambda(lambda x: x.float().div(255.0))

    if config.inputs.video_input.random_aug:
        aug_transform = transforms.RandAugment()
    else:
        aug_transform = transforms.Lambda(lambda x: x)

    train_transform = transforms.Compose(
        [
            aug_transform,
            transforms.RandomResizedCrop(
                config.inputs.image_res,
                scale=(0.5, 1.0),
                interpolation=InterpolationMode.BICUBIC,
            ),
            transforms.RandomHorizontalFlip(),
            type_transform,
            normalize,
        ]
    )

    return train_transform

def get_test_transform(config, test_file):
    vision_enc_name = config.model.vision_encoder.name
    if "internvideo" in vision_enc_name or "vit" in vision_enc_name or "umt" in vision_enc_name:
        mean = (0.485, 0.456, 0.406)
        std = (0.229, 0.224, 0.225)
    elif "clip" in vision_enc_name:
        mean = (0.48145466, 0.4578275, 0.40821073)
        std = (0.26862954, 0.26130258, 0.27577711)
    else:
        raise NotImplementedError(vision_enc_name)

    normalize = transforms.Normalize(mean, std)

    # loaded images and videos are torch.Tensor of torch.uint8 format,
    # ordered as (T, 1 or 3, H, W) where T=1 for image
    type_transform = transforms.Lambda(lambda x: x.float().div(255.0))

    test_transform = transforms.Compose(
        [
            transforms.Resize(
                (config.inputs.image_res, config.inputs.image_res),
                interpolation=InterpolationMode.BICUBIC,
            ),
            type_transform,
            normalize,
        ]
    )
    return test_transform


def create_dataset(dataset_type, config):
    ##########################################################
    # Shared setting for all datasets
    if config.inputs.get('video_input', None) is not None:
        video_reader_type = config.inputs.video_input.get("video_reader_type", "decord")
        video_only_dataset_kwargs_train = dict(
            video_reader_type=video_reader_type,
            sample_type=config.inputs.video_input.sample_type,
            num_frames=config.inputs.video_input.num_frames,
            num_tries=10,  # false tolerance
        )
        video_only_dataset_kwargs_eval = dict(
            video_reader_type=video_reader_type,
            sample_type=config.inputs.video_input.sample_type_test,
            num_frames=config.inputs.video_input.num_frames_test,
            num_tries=1,  # we want to have predictions for all videos
        )
    else:
        logger.warn("Make sure that you don't need video input!!!")
    if config.inputs.get('audio_input', None) is not None:
        audio_reader_type = config.inputs.audio_input.get("audio_reader_type", "torchaudio")
        audio_only_dataset_kwargs_train = dict(
            audio_reader_type=audio_reader_type,
            audio_sample_rate=config.inputs.audio_input.get('audio_sample_rate', 16000),
            max_audio_length=config.inputs.audio_input.get('max_audio_length', 10),
            num_tries=10,
        )
        audio_only_dataset_kwargs_eval = dict(
            audio_reader_type=audio_reader_type,
            audio_sample_rate=config.inputs.audio_input.get('audio_sample_rate_test', 16000),
            max_audio_length=config.inputs.audio_input.get('max_audio_length', 10),
            num_tries=1,
        )
    else:
        logger.warn("Make sure that you don't need audio input!!!")


    if dataset_type == "pt_train":
        # convert to list of lists
        train_files = (
            [config.train_file] if isinstance(config.train_file, dict) else config.train_file
        )
        train_media_types = sorted(list({get_media_type(e) for e in train_files}))

        train_datasets = []
        for m in train_media_types:
            
            # dataset of the same media_type will be mixed in a single Dataset object
            _train_files = [e for e in train_files if get_media_type(e) == m]

            datasets = []
            sample_weights = []
            for train_file in _train_files:
                dataset_cls = get_dataset_cls(dataset_type=dataset_type, media_type=m, data_cfg=train_file)
                if m == "audio":
                    train_transform = None
                else:
                    train_transform = get_train_transform(config, train_file)
                dataset_kwargs = dict(
                    ann_file=train_file,
                    transform=train_transform,
                    num_epochs=config.scheduler.epochs)

                if m == "audio_video":
                    dataset_kwargs.update(video_only_dataset_kwargs_train)
                    dataset_kwargs.update(audio_only_dataset_kwargs_train)
                elif m == "video":
                    dataset_kwargs.update(video_only_dataset_kwargs_train)
                elif m == 'audio':
                    dataset_kwargs.update(audio_only_dataset_kwargs_train)
                elif m != 'image':
                    raise NotImplementedError(m)
                logger.info(f"dataset_type={dataset_type}, train_file={train_file}")
                logger.info(dataset_kwargs)
                logger.info('train_transform:')
                logger.info(str(train_transform))

                datasets.append(dataset_cls(**dataset_kwargs))
                sample_weights.append(train_file.get("sample_weight", 1))
                # assert train_file.get("sample_weight", 1) == 1, train_file

            if sum(sample_weights) > len(sample_weights):
                logger.info(f'Use ResampleConcatDataset for {m}, sample_weights={sample_weights}')
                dataset = ResampleConcatDataset(datasets, sample_weights=sample_weights)
            else:
                logger.info(f'Use ConcatDataset for {m}')
                dataset = ConcatDataset(datasets)

            train_datasets.append(dataset)

        return train_datasets

    elif dataset_type == "ret_train":
        assert isinstance(config.train_file, dict), config.train_file
        train_transform = get_train_transform(config, config.train_file)
        dataset_cls = get_dataset_cls(dataset_type=dataset_type,
                                      media_type=config.train_file.media_type,
                                      data_cfg=config.train_file)
        if config.train_file.media_type == "video":
            dataset_kwargs = dict(
                ann_file=config.train_file,
                transform=train_transform)
            dataset_kwargs.update(video_only_dataset_kwargs_train)
        elif config.train_file.media_type == 'audio':
            dataset_kwargs = dict(
                ann_file=config.train_file,
                transform=None)
            dataset_kwargs.update(audio_only_dataset_kwargs_train)
        elif config.train_file.media_type == 'audio_video':
            dataset_kwargs = dict(
                ann_file=config.train_file,
                transform=train_transform)
            dataset_kwargs.update(video_only_dataset_kwargs_train)
            dataset_kwargs.update(audio_only_dataset_kwargs_train)
        else:
            raise NotImplementedError(config.train_file.media_type)
        
        logger.info(f"dataset_type={dataset_type}, train_file={config.train_file}")
        logger.info(dataset_kwargs)
        logger.info('train_transform:')
        logger.info(str(train_transform))

        return [dataset_cls(**dataset_kwargs)]
    
    elif dataset_type == "qa_train":
        assert type(config.train_file) is dict, f"assuming single train media type but get {config.train_file}"
        
        media_type = get_media_type(config.train_file[0])  # assuming single train media type
        if media_type == "audio":
            train_transform = None
        else:
            train_transform = get_train_transform(config, config.train_file)
            
        dataset_cls = get_dataset_cls(dataset_type=dataset_type,
                                      media_type=media_type,
                                      data_cfg=config.train_file)
        dataset_kwargs = dict(
            ann_file=config.train_file, transform=train_transform, eos=config.eos, mode="train"
        )
        if media_type == "video":
            dataset_kwargs.update(video_only_dataset_kwargs_train)
        train_dataset = dataset_cls(**dataset_kwargs)

        logger.info(f"dataset_type={dataset_type}, train_file={config.train_file}")
        logger.info(dataset_kwargs)
        logger.info('train_transform:')
        logger.info(str(train_transform))

        return train_dataset
    
    elif dataset_type in ["pt_eval", "ret_eval", "qa_eval"]:
        test_datasets = []
        test_dataset_names = []
        # multiple test datasets, all separate
        for name, data_cfg in config.test_file.items():
            media_type = get_media_type(data_cfg)
            test_dataset_names.append(name)
            test_transform = get_test_transform(config, data_cfg)

            if dataset_type == "qa_eval" or (dataset_type == "pt_eval" and "_qa_" in name):
                test_dataset_cls = get_dataset_cls(dataset_type='qa_eval',
                                                   media_type=media_type,
                                                   data_cfg=data_cfg)
                dataset_kwargs = dict(
                    ann_file=data_cfg,
                    transform=test_transform,
                    eos=config.eos,
                    mode="eval",
                    answer_list=config.answer_list,
                )
                if media_type == "video":
                    dataset_kwargs.update(video_only_dataset_kwargs_eval)
                else:
                    raise NotImplementedError(f"media_type={media_type}")
            else: # ret
                test_dataset_cls = get_dataset_cls(dataset_type='ret_eval',
                                                   media_type=media_type,
                                                   data_cfg=data_cfg)
                if media_type == "video":
                    dataset_kwargs = dict(
                        ann_file=data_cfg,
                        transform=test_transform
                    )
                    dataset_kwargs.update(video_only_dataset_kwargs_eval)
                    if 'frame' in data_cfg['data_root']: # read image for video
                        dataset_kwargs["video_reader_type"] = 'img'
                elif media_type == "audio":
                    dataset_kwargs = dict(
                        ann_file=data_cfg,
                        transform=None)
                    dataset_kwargs.update(audio_only_dataset_kwargs_eval)
                elif media_type == 'audio_video':
                    dataset_kwargs = dict(
                        ann_file=data_cfg,
                        transform=test_transform)
                    dataset_kwargs.update(video_only_dataset_kwargs_eval)
                    dataset_kwargs.update(audio_only_dataset_kwargs_eval)
                elif media_type != 'image':
                    raise NotImplementedError(f"media_type={media_type}")
                
            logger.info(f"dataset_type={dataset_type}, test_file={data_cfg}")
            logger.info(dataset_kwargs)
            logger.info('test_transform:')
            logger.info(str(test_transform))

            test_datasets.append(test_dataset_cls(**dataset_kwargs))
        return test_datasets, test_dataset_names
    
    elif dataset_type == "mc_test":
        test_transform = get_test_transform(config, config.test_file.mc_test)
        dataset_kwargs = dict(ann_file=[config.test_file.mc_test], transform=test_transform)
        dataset_kwargs.update(video_only_dataset_kwargs_eval)

        logger.info(f"dataset_type={dataset_type}, test_file={config.test_file}")
        logger.info(dataset_kwargs)
        logger.info('test_transform:')
        logger.info(str(test_transform))
        
        return VidTxtRetMCEvalDataset(**dataset_kwargs)
    
    elif dataset_type == "mc_new_test":
        test_transform = get_test_transform(config, config.test_file.mc_test)
        dataset_kwargs = dict(ann_file=[config.test_file.mc_test], transform=test_transform)
        dataset_kwargs.update(video_only_dataset_kwargs_eval)

        logger.info(f"dataset_type={dataset_type}, test_file={config.test_file}")
        logger.info(dataset_kwargs)
        logger.info('test_transform:')
        logger.info(str(test_transform))
        
        return VidTxtRetMCNewEvalDataset(**dataset_kwargs)
    
    else:
        raise NotImplementedError(f"dataset_type={dataset_type}")

def vqa_collate_fn(batch):
    image_list, question_list, answer_list, weight_list, n = [], [], [], [], []
    for image, question, answer, weights in batch:
        image_list.append(image)
        question_list.append(question)
        weight_list += weights
        answer_list += answer
        n.append(len(answer))
    return (
        torch.stack(image_list, dim=0),
        question_list,
        answer_list,
        torch.Tensor(weight_list),
        n,
    )


def create_sampler(datasets, shuffles, num_tasks, global_rank):
    samplers = []
    for dataset, shuffle in zip(datasets, shuffles):
        sampler = torch.utils.data.DistributedSampler(
            dataset, num_replicas=num_tasks, rank=global_rank, shuffle=shuffle
        )
        samplers.append(sampler)
    return samplers

def create_stateful_sampler(datasets, batch_size):
    samplers = []
    for dataset, bs in zip(datasets, batch_size):
        sampler = StatefulDistributedSampler(dataset, batch_size=bs)
        samplers.append(sampler)
    return samplers

def create_loader(datasets, samplers, batch_size, num_workers, is_trains, collate_fns):
    loaders = []
    for dataset, sampler, bs, n_worker, is_train, collate_fn in zip(
        datasets, samplers, batch_size, num_workers, is_trains, collate_fns
    ):
        if is_train:
            shuffle = sampler is None
            drop_last = True
            pin_memory = True
            persistent_workers = True if n_worker > 0 else False
        else:
            shuffle = False
            drop_last = False
            pin_memory = False
            persistent_workers = False
        loader = DataLoader(
            dataset,
            batch_size=bs,
            num_workers=n_worker,
            pin_memory=pin_memory,
            sampler=sampler,
            shuffle=shuffle,
            collate_fn=collate_fn,
            drop_last=drop_last,
            persistent_workers=persistent_workers,
        )
        loaders.append(loader)
    return loaders


def iterate_dataloaders(dataloaders):
    """Alternatively generate data from multiple dataloaders,
    since we use `zip` to concat multiple dataloaders,
    the loop will end when the smaller dataloader runs out.

    Args:
        dataloaders List(DataLoader): can be a single or multiple dataloaders
    """
    for data_tuples in zip(*dataloaders):
        for idx, data in enumerate(data_tuples):
            yield dataloaders[idx].dataset.media_type, data



"""Tests for temporal_flickering module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_temporal_flickering_basics():
    from ayase.modules.temporal_flickering import TemporalFlickeringModule
    _test_module_basics(TemporalFlickeringModule, "temporal_flickering")

def test_temporal_flickering_video(video_sample):
    from ayase.modules.temporal_flickering import TemporalFlickeringModule
    video_sample.quality_metrics = QualityMetrics()
    m = TemporalFlickeringModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
Train and eval functions used in main.py
"""
import math
import os
import sys
from typing import Iterable

import torch
import util.misc as utils
from datasets.coco_eval import CocoEvaluator, convert_to_xywh
from datasets.data_prefetcher import data_prefetcher
from datasets.panoptic_eval import PanopticEvaluator
from util.ema import requires_grad, update_ema
from util.misc import NestedTensor


def train_one_epoch(
    model: torch.nn.Module,
    criterion: torch.nn.Module,
    data_loader: Iterable,
    optimizer: torch.optim.Optimizer,
    device: torch.device,
    epoch: int,
    max_norm: float = 0,
    ema: torch.nn.Module = None,
    ema_decay: float = 0.999,
):
    model.train()
    criterion.train()
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter("lr", utils.SmoothedValue(window_size=1, fmt="{value:.6f}"))
    metric_logger.add_meter(
        "class_error", utils.SmoothedValue(window_size=1, fmt="{value:.2f}")
    )
    metric_logger.add_meter(
        "grad_norm", utils.SmoothedValue(window_size=1, fmt="{value:.2f}")
    )
    header = "Epoch: [{}]".format(epoch)
    print_freq = 10

    prefetcher = data_prefetcher(data_loader, device, prefetch=True)
    samples, targets = prefetcher.next()

    # for samples, targets in metric_logger.log_every(data_loader, print_freq, header):
    for _ in metric_logger.log_every(range(len(data_loader)), print_freq, header):
        outputs = model(samples)
        loss_dict = criterion(outputs, targets)
        weight_dict = criterion.weight_dict
        losses = sum(
            loss_dict[k] * weight_dict[k] for k in loss_dict.keys() if k in weight_dict
        )

        # reduce losses over all GPUs for logging purposes
        loss_dict_reduced = utils.reduce_dict(loss_dict)
        loss_dict_reduced_unscaled = {
            f"{k}_unscaled": v for k, v in loss_dict_reduced.items()
        }
        loss_dict_reduced_scaled = {
            k: v * weight_dict[k]
            for k, v in loss_dict_reduced.items()
            if k in weight_dict
        }
        losses_reduced_scaled = sum(loss_dict_reduced_scaled.values())

        loss_value = losses_reduced_scaled.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            print(loss_dict_reduced)
            sys.exit(1)

        optimizer.zero_grad()
        losses.backward()
        if max_norm > 0:
            grad_total_norm = torch.nn.utils.clip_grad_norm_(
                model.parameters(), max_norm
            )
        else:
            grad_total_norm = utils.get_total_grad_norm(model.parameters(), max_norm)
        optimizer.step()

        if ema is not None:
            update_ema(ema, model.module, ema_decay)
        # torch.cuda.empty_cache()

        metric_logger.update(
            loss=loss_value, **loss_dict_reduced_scaled, **loss_dict_reduced_unscaled
        )
        metric_logger.update(class_error=loss_dict_reduced["class_error"])
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])
        metric_logger.update(grad_norm=grad_total_norm)

        samples, targets = prefetcher.next()
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}


@torch.no_grad()
def evaluate(
    model_no_ema,
    criterion,
    postprocessors,
    data_loader,
    base_ds,
    device,
    output_dir,
    test_hflip_aug,
    tta,
    soft_nms,
    ema=None,
    save_result=False,
    save_result_dir="",
    soft_nms_method="quad",
    nms_thresh=0.7,
    quad_scale=0.5,
    lsj_img_size=1824,
):
    model = model_no_ema if ema is None else ema
    model.eval()
    criterion.eval()

    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter(
        "class_error", utils.SmoothedValue(window_size=1, fmt="{value:.2f}")
    )
    header = "Test:"

    iou_types = tuple(k for k in ("segm", "bbox") if k in postprocessors.keys())
    coco_evaluator = CocoEvaluator(base_ds, iou_types)
    # coco_evaluator.coco_eval[iou_types[0]].params.iouThrs = [0, 0.1, 0.5, 0.75]

    panoptic_evaluator = None
    if "panoptic" in postprocessors.keys():
        panoptic_evaluator = PanopticEvaluator(
            data_loader.dataset.ann_file,
            data_loader.dataset.ann_folder,
            output_dir=os.path.join(output_dir, "panoptic_eval"),
        )

    prediction_list = []
    for samples, targets in metric_logger.log_every(data_loader, 10, header):
        samples = samples.to(device)
        targets = [{k: v.to(device) for k, v in t.items()} for t in targets]

        if test_hflip_aug:
            assert (
                samples.tensors.shape[0] == 1
            ), "test_hflip_aug only supports batch size 1"
            assert (
                samples.tensors.shape[1] == 6
            ), "test_hflip_aug requires two images in a batch"
            first_samples = NestedTensor(samples.tensors[:, :3], samples.mask)
            outputs = model(first_samples)
            flipped_samples = NestedTensor(samples.tensors[:, 3:], samples.mask)
            flipped_outputs = model(flipped_samples)
        else:
            outputs = model(samples)
        loss_dict = criterion(outputs, targets)
        weight_dict = criterion.weight_dict

        # reduce losses over all GPUs for logging purposes
        loss_dict_reduced = utils.reduce_dict(loss_dict)
        loss_dict_reduced_scaled = {
            k: v * weight_dict[k]
            for k, v in loss_dict_reduced.items()
            if k in weight_dict
        }
        loss_dict_reduced_unscaled = {
            f"{k}_unscaled": v for k, v in loss_dict_reduced.items()
        }
        metric_logger.update(
            loss=sum(loss_dict_reduced_scaled.values()),
            **loss_dict_reduced_scaled,
            **loss_dict_reduced_unscaled,
        )
        metric_logger.update(class_error=loss_dict_reduced["class_error"])

        orig_target_sizes = torch.stack([t["orig_size"] for t in targets], dim=0)
        if test_hflip_aug:
            new_outputs = {}
            pred_logits = outputs["pred_logits"]
            pred_boxes = outputs["pred_boxes"]

            flipped_pred_logits = flipped_outputs["pred_logits"]
            flipped_pred_boxes = flipped_outputs["pred_boxes"]

            reflipped_pred_boxes = flipped_pred_boxes[
                :, :, [0, 1, 2, 3]
            ] * torch.as_tensor([-1, 1, 1, 1]).to(
                flipped_pred_boxes.device
            ) + torch.as_tensor(
                [1, 0, 0, 0]
            ).to(
                flipped_pred_boxes.device
            )

            new_pred_logits = torch.cat([pred_logits, flipped_pred_logits], dim=1)
            new_pred_boxes = torch.cat([pred_boxes, reflipped_pred_boxes], dim=1)

            new_outputs["pred_logits"] = new_pred_logits
            new_outputs["pred_boxes"] = new_pred_boxes
            results = postprocessors["bbox"](
                new_outputs, 
                orig_target_sizes, 
                soft_nms=soft_nms,
                method=soft_nms_method,
                nms_thresh=nms_thresh,
                quad_scale=quad_scale,
            )
        else:
            results = postprocessors["bbox"](
                outputs, 
                orig_target_sizes, 
                soft_nms=soft_nms,
                method=soft_nms_method,
                nms_thresh=nms_thresh,
                quad_scale=quad_scale,
            )
        if "segm" in postprocessors.keys():
            target_sizes = torch.stack([t["size"] for t in targets], dim=0)
            results = postprocessors["segm"](
                results, outputs, orig_target_sizes, target_sizes
            )
        res = {
            target["image_id"].item(): output
            for target, output in zip(targets, results)
        }
        if coco_evaluator is not None:
            coco_evaluator.update(res)

        if panoptic_evaluator is not None:
            res_pano = postprocessors["panoptic"](
                outputs, target_sizes, orig_target_sizes
            )
            for i, target in enumerate(targets):
                image_id = target["image_id"].item()
                file_name = f"{image_id:012d}.png"
                res_pano[i]["image_id"] = image_id
                res_pano[i]["file_name"] = file_name

            panoptic_evaluator.update(res_pano)

        for target, output in zip(targets, results):
            res_cpu = {
                target["image_id"].item(): {
                    "boxes": output["boxes"].cpu(),
                    "labels": output["labels"].cpu(),
                    "scores": output["scores"].cpu(),
                }
            }
            prediction_list.append(res_cpu)

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)

    if save_result:

        from torch import distributed as dist

        os.makedirs(save_result_dir, exist_ok=True)
        rank = dist.get_rank()
        torch.save(
            prediction_list,
            os.path.join(save_result_dir, f"val2017_prediction_{rank}.pth"),
        )

    if coco_evaluator is not None:
        coco_evaluator.synchronize_between_processes()
    if panoptic_evaluator is not None:
        panoptic_evaluator.synchronize_between_processes()

    # accumulate predictions from all images
    if coco_evaluator is not None:
        coco_evaluator.accumulate()
        coco_evaluator.summarize()
    panoptic_res = None
    if panoptic_evaluator is not None:
        panoptic_res = panoptic_evaluator.summarize()
    stats = {k: meter.global_avg for k, meter in metric_logger.meters.items()}
    if coco_evaluator is not None:
        if "bbox" in postprocessors.keys():
            stats["coco_eval_bbox"] = coco_evaluator.coco_eval["bbox"].stats.tolist()
        if "segm" in postprocessors.keys():
            stats["coco_eval_masks"] = coco_evaluator.coco_eval["segm"].stats.tolist()
    if panoptic_res is not None:
        stats["PQ_all"] = panoptic_res["All"]
        stats["PQ_th"] = panoptic_res["Things"]
        stats["PQ_st"] = panoptic_res["Stuff"]
    return stats, coco_evaluator




# Copyright (c) Meta Platforms, Inc. and affiliates.

import abc
import logging
import os
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Tuple

import tiktoken
from sentencepiece import SentencePieceProcessor
from tiktoken.load import load_tiktoken_bpe

logger = logging.getLogger(__name__)


@dataclass
class TokenizerArgs:
    name: str = "bytes"
    path: Optional[str] = None


class Tokenizer(abc.ABC):
    @abc.abstractmethod
    def encode(self, tokens, add_bos, add_eos):
        pass

    @abc.abstractmethod
    def decode(self, tokens):
        pass

    @abc.abstractmethod
    def get_token_offsets(
        self, text: str, tokens: Optional[List[int]] = None
    ) -> Tuple[List[str], List[int]]:
        """Return the offsets of the tokens in the original text. Only used for evaluation."""
        pass


class MockTokenizer(Tokenizer):
    n_words: int = 256

    def encode(self, tokens, add_bos, add_eos):
        return tokens


class ByteTokenizer(Tokenizer):
    def __init__(self):
        self.bos_id = 256
        self.eos_id = 257
        self.n_words = 258

    def encode(self, s: str, add_bos: bool = False, add_eos: bool = False):
        tokens = [self.bos_id] * add_bos + list(s.encode()) + [self.eos_id] * add_eos
        return tokens

    def decode(self, tokens: List[int]):
        byte_tokens = bytes([t for t in tokens if t < 256])
        return byte_tokens.decode("utf-8", errors="backslashreplace")

    def get_token_offsets(
        self, text: str, tokens: Optional[List[int]] = None
    ) -> Tuple[List[str], List[int]]:
        if tokens is None:
            tokens = self.encode(text)

        decoded_chars, offsets = [], []
        byte_pos = 0
        for token in tokens:
            if token < 256:
                char = bytes([token]).decode("utf-8", errors="ignore")
                if char:
                    decoded_chars.append(char)
                    offsets.append(byte_pos)
                byte_pos += len(char.encode("utf-8"))

        return decoded_chars, offsets


class SentencePieceTokenizer(Tokenizer):
    def __init__(self, model_path: str) -> None:
        assert os.path.isfile(model_path), model_path
        self.sp_model = SentencePieceProcessor(model_file=model_path)

        logger.info(f"Reloaded SentencePiece model from {model_path}")

        # BOS / EOS token IDs
        self.n_words: int = self.sp_model.vocab_size()
        self.bos_id: int = self.sp_model.bos_id()
        self.eos_id: int = self.sp_model.eos_id()
        self.pad_id: int = self.sp_model.pad_id()
        logger.info(
            f"#words: {self.n_words} - BOS ID: {self.bos_id} - EOS ID: {self.eos_id}"
        )
        assert self.sp_model.vocab_size() == self.sp_model.get_piece_size()

    def encode(self, s: str, add_bos: bool, add_eos: bool):
        assert type(s) is str
        tokens = (
            [self.bos_id] * add_bos + self.sp_model.encode(s) + [self.eos_id] * add_eos
        )
        return tokens

    def decode(self, tokens: List[int]):
        return self.sp_model.decode(tokens)

    def get_token_offsets(
        self, text: str, tokens: Optional[List[int]] = None
    ) -> Tuple[List[str], List[int]]:
        pieces = self.sp_model.encode_as_immutable_proto(text).pieces
        substrs = [p.surface for p in pieces]
        offsets = [p.begin for p in pieces]
        return substrs, offsets


DEFAULT_TIKTOKEN_PATTERN = r"""(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"""
DEFAULT_TIKTOKEN_SPECIAL_TOKENS = {
    "<|begin_of_text|>": 0,
    "<|end_of_text|>": 1,
    "<|fim_prefix|>": 2,
    "<|fim_middle|>": 3,
    "<|fim_end_fill|>": 253,
    "<|fim_pad|>": 254,
    "<|fim_suffix|>": 255,
}
TIKTOKEN_MAX_ENCODE_CHARS = 400_000


class TikTokenTokenizer(Tokenizer):

    def __init__(self, model_path: str) -> None:
        mergeable_ranks = load_tiktoken_bpe(model_path)
        all_special_tokens_with_ids = copy(DEFAULT_TIKTOKEN_SPECIAL_TOKENS)
        missing_ids = set(range(256)) - set(all_special_tokens_with_ids.values())
        for id in missing_ids:
            all_special_tokens_with_ids[f"<|reserved_special_token_{id}|>"] = id
        for name in all_special_tokens_with_ids:
            all_special_tokens_with_ids[name] += len(mergeable_ranks)

        self.tkt_model = tiktoken.core.Encoding(
            name=Path(model_path).stem,
            pat_str=DEFAULT_TIKTOKEN_PATTERN,
            mergeable_ranks=mergeable_ranks,
            special_tokens=all_special_tokens_with_ids,
        )

        self.bos_id: int = self.tkt_model.encode_single_token("<|begin_of_text|>")
        self.eos_id: int = self.tkt_model.encode_single_token("<|end_of_text|>")

        self.n_words: int = self.tkt_model.n_vocab

        logger.info(
            f"#words: {self.n_words} - BOS ID: {self.bos_id} - EOS ID: {self.eos_id}"
        )

    def encode(self, s: str, add_bos: bool, add_eos: bool):
        assert isinstance(s, str)

        subs = []
        for i in range(0, len(s), TIKTOKEN_MAX_ENCODE_CHARS):
            subs.append(s[i : i + TIKTOKEN_MAX_ENCODE_CHARS])
        return (
            [self.bos_id] * add_bos
            + sum(self.tkt_model.encode_ordinary_batch(subs), start=[])
            + [self.eos_id] * add_eos
        )

    def decode(self, tokens: List[int]):
        return self.tkt_model.decode(tokens)

    def get_token_offsets(
        self, text: str, tokens: Optional[List[int]] = None
    ) -> Tuple[List[str], List[int]]:
        if tokens is not None:
            token_bytes = self.tkt_model.decode_tokens_bytes(tokens)
        else:
            token_bytes = self.tkt_model.decode_tokens_bytes(
                self.tkt_model.encode(text, allowed_special="all")
            )

        text_len, offsets = 0, []
        for token in token_bytes:
            offsets.append(max(0, text_len - (0x80 <= token[0] < 0xC0)))
            text_len += sum(1 for c in token if not 0x80 <= c < 0xC0)
        substrs = [text[s:e] for s, e in zip(offsets, offsets[1:] + [None])]
        return substrs, offsets


def build_tokenizer(name: str, path: Optional[str] = None) -> Tokenizer:
    if name == "bytes":
        return ByteTokenizer()
    elif name == "mock":
        return MockTokenizer()
    elif name == "sp":
        return SentencePieceTokenizer(path)
    elif name == "tiktoken":
        return TikTokenTokenizer(path)
    else:
        raise NotImplementedError(f"{name} tokenizer type is not implemented")



import numpy as np
import cv2
from joblib import dump,Parallel,delayed
from skimage.util.shape import view_as_blocks,view_as_windows
from skimage.util import apply_parallel

TV_HEIGHT = 32.6
VIEWING_DIST = 1.5*TV_HEIGHT
theta = TV_HEIGHT/2160*180/np.pi*1/VIEWING_DIST
f_max = 1/theta
h_win = 45 
w_win = 45 
X =h_win/2160*1/1.5
def gen_gauss_window(l=5, sig=1.):
    """\
    creates gaussian kernel with side length l and a sigma of sig
    """

    ax = np.linspace(-(l - 1) / 2., (l - 1) / 2., int(l))
    xx, yy = np.meshgrid(ax, ax)

    kernel = np.exp(-0.5 * (np.square(xx) + np.square(yy)) / np.square(sig))

    return kernel / np.sum(kernel)
gauss_window = gen_gauss_window(h_win,h_win/6)


def csf_barten_frequency(f,L,X=40):
    '''
    Implements simplified CSF from Peter Barten,"Formula for the contrast sensitivity of the human eye", Electronic Imaging 2004
    '''
    
    num = 5200*np.exp(-0.0016*(f**2)*((1+100/L)**0.08))
    denom1 = 1+144/X**2+0.64*f**2
    denom21 = 63/L**0.83
    denom22 = 1/(1e-4+1-np.exp(-0.02*(f**2)))
    denom2 =denom21+denom22 
    csf_freq = num/(np.sqrt(denom1*denom2))
    return csf_freq
def round_down(num,divisor):
    return num-(num%divisor)

def blockwise_csf(y,adaptation='gaussian',h_win=h_win,w_win=w_win):
    '''
    Divides frame into non-overlapping blocks and applies Barten's CSF
    '''

    max_h,max_w = round_down(y.shape[0],h_win),round_down(y.shape[1],w_win)
    y_crop = y[:max_h,:max_w]
    blocks_list = view_as_blocks(y_crop,(h_win,w_win))
    blocks =np.reshape(blocks_list,(-1,blocks_list.shape[2],blocks_list.shape[3]))
    block_csf = Parallel(n_jobs=-5,verbose=0)(delayed(csf_filter_block)(block=block,use_index=False,adaptation=adaptation,overlap=False) for block in blocks)
    block_csf = np.reshape(block_csf,(max_h,max_w))

    return block_csf

def windows_csf(y,use_views=False,adaptation='bilateral',h_win=h_win,w_win=w_win):
    '''
    Divides frame into overlapping blocks and applies Barten's CSF
    '''

 #TODO find out if using views is faster
    if(use_views): 
        max_h,max_w = round_down(y.shape[0],h_win),round_down(y.shape[1],w_win)
        y_crop = y[:max_h,:max_w]
        #window_csf = apply_parallel(csf_filter_block,y_crop,(h_win,w_win),extra_keywords={"window":True},dtype=np.float32)
        windows_list = view_as_windows(y_crop,(h_win,w_win))
        out_h,out_w = windows_list.shape[0],windows_list.shape[1]
        windows =np.reshape(windows_list,(-1,windows_list.shape[2],windows_list.shape[3]))
        window_csf = Parallel(n_jobs=-5,verbose=0)(delayed(csf_filter_block)(window,overlap=True) for window in windows)
    else:
        h_indices = np.arange(h_win//2,y.shape[0]-h_win//2)
        w_indices = np.arange(w_win//2,y.shape[1]-w_win//2)
        xx,yy = np.meshgrid(h_indices,w_indices)
        xx = xx.flatten()
        yy = yy.flatten()

        window_csf = Parallel(n_jobs=-5,verbose=0)(delayed(csf_filter_block)(y=y,adaptation=adaptation,index=pixel_index,xx_list=xx,yy_list=yy,h_win=h_win,w_win=w_win,use_index=True,overlap=True) for pixel_index in range(len(xx)))
        out_h,out_w = len(h_indices),len(w_indices)
    window_csf = np.reshape(window_csf,(out_h,out_w))
    return window_csf



def csf_filter_block(y=None,block=None,adaptation='gaussian',index=None,xx_list=None,yy_list=None,h_win=None,w_win=None,use_index=True,overlap=True,pq=True,gauss_window=gauss_window,X=X):
    if(use_index):
        xx,yy = xx_list[index],yy_list[index]
        block = y[xx-h_win//2:xx+h_win//2+1, yy-w_win//2:yy+w_win//2+1]
    block = block.astype(np.float32)
    if(adaptation=='gaussian'):
        gauss_filtered = gauss_window*block
        avg_luminance = np.sum(gauss_filtered)
    elif(adaptation=='bilateral'):
        bilateral_filtered = cv2.bilateralFilter(src=block,d=-1,sigmaColor=20,sigmaSpace=20)
        avg_luminance = np.average(bilateral_filtered)


    h,w = block.shape
    u_min = -(h >> 1)
    u_max = (h >> 1) + 1 if h & 1 else (h >> 1)
    v_min = -(w >> 1)
    v_max = (w >> 1) + 1 if w & 1 else (w >> 1)

    u, v = np.meshgrid(np.arange(u_min, u_max), np.arange(v_min, v_max), indexing='ij')
    fx, fy = u*f_max/h, v*f_max/w

    csf_mat = csf_barten_frequency(np.abs(fx),avg_luminance,X) * csf_barten_frequency(np.abs(fy),avg_luminance,X)
    if(pq==True):
        csf_mat = csf_mat/np.max(csf_mat)

    block_csf_filtered = np.real(np.fft.ifft2(np.fft.ifftshift(np.fft.fftshift(np.fft.fft2(block)) * csf_mat)))
    if(overlap):
       return np.reshape(block_csf_filtered[h//2,w//2],(-1,1))  
    else:
        return block_csf_filtered


def csf_frequency(f):
    return (0.31 + 0.69*f) * np.exp(-0.29*f)


def csf_spat_filter(d2h, k=21):
    assert isinstance(k, int) and k > 0 and (k & 1), 'The length of the filter must be an odd positive integer'
    del_theta = 180 / (d2h * 1080 * np.pi)
    t = np.arange(-(k >> 1), (k >> 1) + 1) * del_theta
    assert len(t) == k, 'Filter is of the wrong size'

    a = 0.31
    b = 0.69
    c = 0.29
    f = 2*((a*c + b)*c**2 + (a*c - b)*4*np.pi**2 * t**2) / (c**2 + 4*np.pi**2 * t**2)**2   # Inverse Fourier Transform of CSF.

    return f*del_theta


# Ref "Most apparent distortion: full-reference image quality assessment and the role of strategy", E. C. Larson and D. M. Chandler
def csf_mannos_daly(f, theta):
    f_theta = f / (0.15*np.cos(theta) + 0.85)
    lamda = 0.228
    f_max = 4  # TODO: Find out how f_max is obtained from lamda
    if isinstance(f, np.ndarray):
        return np.where(f >= f_max, 2.6*(0.0192 + lamda*f_theta)*np.exp(-np.power(lamda*f_theta, 1.1)), 0.981)
    else:
        return 2.6*(0.0192 + lamda*f_theta)*np.exp(-np.power(lamda*f_theta, 1.1)) if f >= f_max else 0.981


def detection_threshold(a, k, f0, g, level_frequency):
    return a * np.power(10, k*np.log10(f0 * g / level_frequency)**2)


def csf_adm(level, subband):
    '''
    level: 0 indexed level of the discrete wavelet transform
    subband: 0 indexed in the list [approximation, horizontal, vertical, diagonal]
    '''
    # Distance to height ratio of the display
    d2h = 3.0
    pic_height = 1080
    factor = np.pi*pic_height*d2h/180
    level_frequency = factor / (1 << (level+1))
    orientation_factors = [1.0/0.85, 1.0, 1.0, 1/(0.85-0.15)]
    return csf_frequency(level_frequency * orientation_factors[subband])


def csf_cdf97_watson(level, subband):
    '''
    level: 0 indexed level of the discrete wavelet transform
    subband: 0 indexed in the list [approximation, horizontal, vertical, diagonal]
    Ref: A. Watson, G. Yang, et al. "Visibility of Wavelet Quantization Noise"
    '''
    # Detection threshold model parameters
    a = 0.495
    k = 0.466
    f0 = 0.401
    gs = [1.501, 1, 1, 0.534]

    # Distance to height ratio of the display
    d2h = 3.0
    pic_height = 1080
    factor = np.pi*pic_height*d2h/180
    level_frequency = factor / (1 << (level+1))

    # Basis function amplitudes
    amplitudes = np.array([[0.621710, 0.672340, 0.672340, 0.727090],
                           [0.345370, 0.413170, 0.413170, 0.494280],
                           [0.180040, 0.227270, 0.227270, 0.286880],
                           [0.091401, 0.117920, 0.117920, 0.152140],
                           [0.045943, 0.059758, 0.059758, 0.077727],
                           [0.023013, 0.030018, 0.030018, 0.039156]])

    return 0.5 * amplitudes[level, subband] / detection_threshold(a, k, f0, gs[subband], level_frequency)


def csf_dwt_hill(level, subband):
    '''
    level: 0 indexed level of the discrete wavelet transform
    subband: 0 indexed in the list [approximation, horizontal, vertical, diagonal]
    Ref: P. Hill, A. Achim, et al. "Contrast Sensitivity of the Wavelet, Dual Tree Complex Wavelet, Curvelet, and Steerable Pyramid Transforms"
    '''
    # Detection threshold model parameters
    a = 2.818
    k = 0.783
    f0 = 0.578
    gs = [1.5, 1, 1, 0.534]  # g0, i.e. for approximation subband, is not provided. Using value from Watson (ref: csf_cdf97_watson). Do not recommend using.

    # Distance to height ratio of the display
    d2h = 3.0
    pic_height = 1080
    factor = np.pi*pic_height*d2h/180
    level_frequency = factor / (1 << (level+1))

    # Basis function amplitudes
    # amplitudes = np.array([[0.621710, 0.672340, 0.672340, 0.727090],
    #                        [0.345370, 0.413170, 0.413170, 0.494280],
    #                        [0.180040, 0.227270, 0.227270, 0.286880],
    #                        [0.091401, 0.117920, 0.117920, 0.152140],
    #                        [0.045943, 0.059758, 0.059758, 0.077727],
    #                        [0.023013, 0.030018, 0.030018, 0.039156]])

    # return 2.0 * detection_threshold(a, k, f0, gs[subband], level_frequency) / amplitudes[level, subband]
    return 1.0 / (20 * np.log10(detection_threshold(a, k, f0, gs[subband], level_frequency) / 128))


def csf_dtcwt_hill(level, subband):
    '''
    level: 0 indexed level of the discrete wavelet transform
    subband: 0 indexed in the list [approximation, horizontal, vertical, diagonal]
    Ref: P. Hill, A. Achim, et al. "Contrast Sensitivity of the Wavelet, Dual Tree Complex Wavelet, Curvelet, and Steerable Pyramid Transforms"
    '''
    # Detection threshold model parameters
    a = 3.107
    k = 1.025
    f0 = 0.755
    gs = [1.3, 1, 1, 0.814]  # g0, i.e. for approximation subband, is not provided. Guessing. Do not use recommend using.

    # Distance to height ratio of the display
    d2h = 3.0
    pic_height = 1080
    factor = np.pi*pic_height*d2h/180
    level_frequency = factor / (1 << (level+1))

    return 1.0 / (20 * np.log10(detection_threshold(a, k, f0, gs[subband], level_frequency) / 128))


def csf_curvelet_hill(level, subband):
    '''
    level: 0 indexed level of the discrete wavelet transform
    subband: 0 indexed in the list [approximation, horizontal, vertical, diagonal]
    Ref: P. Hill, A. Achim, et al. "Contrast Sensitivity of the Wavelet, Dual Tree Complex Wavelet, Curvelet, and Steerable Pyramid Transforms"
    '''
    # Detection threshold model parameters
    a = 1.083
    k = 0.790
    f0 = 0.509
    gs = [1, 1, 1, 1]  # g0, i.e. for approximation subband, is not provided. Guessing. Do not use recommend using.

    # Distance to height ratio of the display
    d2h = 3.0
    pic_height = 1080
    factor = np.pi*pic_height*d2h/180
    level_frequency = factor / (1 << (level+1))

    return 1.0 / (20 * np.log10(detection_threshold(a, k, f0, gs[subband], level_frequency) / 128))


def csf_steerable_hill(level, subband):
    '''
    level: 0 indexed level of the discrete wavelet transform
    subband: 0 indexed in the list [approximation, horizontal, vertical, diagonal]
    Ref: P. Hill, A. Achim, et al. "Contrast Sensitivity of the Wavelet, Dual Tree Complex Wavelet, Curvelet, and Steerable Pyramid Transforms"
    '''
    # Detection threshold model parameters
    a = 2.617
    k = 0.960
    f0 = 0.487
    gs = [1, 1, 1, 1]  # g0, i.e. for approximation subband, is not provided. Guessing. Do not use recommend using.

    # Distance to height ratio of the display
    d2h = 3.0
    pic_height = 1080
    factor = np.pi*pic_height*d2h/180
    level_frequency = factor / (1 << (level+1))

    return 1.0 / (20 * np.log10(detection_threshold(a, k, f0, gs[subband], level_frequency) / 128))


def ahc_weight(level, subband, n_levels, binarized=True):
    '''
    Weighting function used in Adaptive High Frequency Clipping
    If binarized is True, weight is compared to a threshold and 0-1 outputs are returned.
    Ref: K. Gu, G. Zhai, et al. "Adaptive High Frequency Clipping for Improved Image Quality Assessment"
    '''
    # Weighting function parameters
    a = 10
    k = 10
    t = 2
    d0 = 512
    gs = [1, 2, 2, 1]  # g0, i.e. for approximation subband, is not provided. Guessing. Do not use recommend using.
    thresh = 1.0

    # Distance to height ratio of the display
    d2h = 3.0

    weight = gs[subband] * np.power(k, t*(n_levels - (level+1))) / np.power(a,  d2h/d0)  # Paper says "d/d0" but I think they meant to use d2h
    if binarized:
        weight = float(weight >= thresh)

    return weight


csf_dict = {'frequency': csf_frequency,
            'spat_filter': csf_spat_filter,
            'adm': csf_adm,
            'mannos_daly': csf_mannos_daly,
            'cdf97_watson': csf_cdf97_watson,
            'dwt_hill': csf_dwt_hill,
            'dtcwt_hill': csf_dtcwt_hill,
            'curvelet_hill': csf_curvelet_hill,
            'steerable_hill': csf_steerable_hill,
            'ahc': ahc_weight}



"""Datamodule for Llava Pretraining and Finetuning"""
import os
import re
from PIL import Image
import numpy as np
import re
import tempfile
from typing import Dict, List, Union, Tuple
import traceback
import json

import torch
import torch.nn.functional as F
from transformers import DataCollatorForSeq2Seq

from ..tools.rw_utils import read_jsonlines
from torch.utils.data import Dataset, DataLoader

np_str_obj_array_pattern = re.compile(r"[SaUO]")

default_collate_err_msg_format = (
    "default_collate: batch must contain tensors, numpy arrays, numbers, "
    "dicts or lists; found {}"
)

from .custom_data_parsers.standard_vision_parser import VisionParser
from .custom_data_parsers.object_tracking_parser import ObjectTrackingParser
from .custom_data_parsers.multi_images_parser import MultiImagesParser
from .custom_data_parsers.video_permutation_parser import VideoPermutationParser
from .custom_data_parsers.utils_visualize import visualize_image_bbox

from .tarsier_processor import TarsierProcessor

from ..tools.rw_utils import NumpyArrayEncoder
from .utils import DictToObject

class TarsierDataProcessor:
    def __init__(
        self,
        processor: TarsierProcessor,
        n_frames: Union[int, list],
        max_n_frames=256,
        max_pixels=int(1280 * 720 // 2),
        min_pixels=0,
        max_seq_len=None,
        is_training=True,  # 会影响：1. 训练和测试时采帧不同；2. 测试时忽略 response。
        print_data_error=True,
        do_image_padding=False,
        do_image_crop=False,
        do_image_resize=True,
        video_sampling_strategy={},
        prompt='',
        train_task='sft',
        **kwargs
    ):
        self.kwargs = kwargs

        self.processor = processor
        self.pad_collator = DataCollatorForSeq2Seq(processor.tokenizer, padding='longest')
        
        self.processor.max_seq_len = self.tokenizer.model_max_length if max_seq_len is None else max_seq_len

        self.n_frames = n_frames
        self.max_n_frames = max_n_frames
        self.max_pixels = max_pixels
        self.min_pixels = min_pixels

        self.is_training = is_training
        self.print_data_error = print_data_error
        self.do_image_padding = do_image_padding
        self.do_image_crop = do_image_crop
        self.do_image_resize = do_image_resize
        self.video_sampling_strategy = video_sampling_strategy
        self.prompt = prompt
        self.train_task = train_task

        self.object_tracking_parser = ObjectTrackingParser(
            n_frames=self.n_frames,
            max_objects=4,
            is_training=self.is_training,
        )
        self.multi_images_parser = MultiImagesParser(
            n_frames=self.n_frames,
            is_training=self.is_training,
        )
        self.video_permutation_parser = VideoPermutationParser(
            n_frames=self.n_frames,
            is_training=self.is_training,
            video_sampling_strategy=self.video_sampling_strategy,
        )
        self.vision_parser = VisionParser(
            n_frames=self.n_frames,
            max_n_frames=self.max_n_frames,
            is_training=self.is_training,
            video_sampling_strategy=self.video_sampling_strategy
        )

    def select_parser(self, data_dict):
        if data_dict.get('task', None) == 'video/object_tracking':
            return self.object_tracking_parser
        elif data_dict.get('task', None) == 'multi_images':
            return self.multi_images_parser
        elif data_dict.get('dataset', None) == 'video_permutation':
            return self.video_permutation_parser
        else:
            return self.vision_parser
    
    def parse_image_processing_config(self, data_dict):
        image_processing_config=data_dict.get('image_processing_config', {})

        do_padding = image_processing_config.get('do_padding', self.do_image_padding)
        do_crop = image_processing_config.get('do_crop', self.do_image_crop)
        do_resize = image_processing_config.get('do_resize', self.do_image_resize)
        max_pixels = image_processing_config.get('max_pixels', self.max_pixels)
        min_pixels = image_processing_config.get('min_pixels', self.min_pixels)

        assert min_pixels <= max_pixels

        image_processing_config['do_padding'] = do_padding
        image_processing_config['do_crop'] = do_crop
        image_processing_config['do_resize'] = do_resize
        image_processing_config['max_pixels'] = max_pixels
        image_processing_config['min_pixels'] = min_pixels

        return image_processing_config
            

    def _transform(self, raw_data_dict: Dict) -> Dict:
        data_dict = json.loads(json.dumps(raw_data_dict, cls=NumpyArrayEncoder))
        del raw_data_dict

        if self.prompt:
            for msg in data_dict['messages']:
                if msg['role'] == 'user':
                    for content in msg['content']:
                        if content['type'] == 'text':
                            content['text'] = self.prompt

        data_dict_copy = json.loads(json.dumps(data_dict, cls=NumpyArrayEncoder))

        image_processing_config = self.parse_image_processing_config(data_dict)
        parser = self.select_parser(data_dict)
        messages = parser.transform(data_dict, image_processing_config)
        data_dict_copy['extra_info'] = data_dict.pop('extra_info', {})

        # visualize_image_bbox(data_dict, image_processing_config, self.processor)
        outputs = self.processor(messages, image_processing_config, is_training=self.is_training)
        
        # if not self.is_training:
        outputs['raw_data_dict'] = data_dict_copy

        return [outputs]
    
    def _split_chosen_rejected(self, data_dict: Dict):
        chosen_data_dict = data_dict
        rejected_data_dict = json.loads(json.dumps(data_dict, cls=NumpyArrayEncoder))
        for msg in chosen_data_dict['messages']:
            if msg['role'] == 'assistant':
                for content in msg['content']:
                    if content['type'] == 'text':
                        content['text'] = content['chosen']
        
        for msg in rejected_data_dict['messages']:
            if msg['role'] == 'assistant':
                for content in msg['content']:
                    if content['type'] == 'text':
                        content['text'] = content['rejected']

        return chosen_data_dict, rejected_data_dict

    def transform(self, data_dict: Dict) -> Dict:
        try:
            if self.train_task == 'dpo':
                chosen_data_dict, rejected_data_dict = self._split_chosen_rejected(data_dict)
                return self._transform(chosen_data_dict) + self._transform(rejected_data_dict)
            return self._transform(data_dict)
        except Exception as e:
            if self.print_data_error:
                print(traceback.format_exc())
                print(f'Error occurs when processing: \n{data_dict}')
            return []

    def batch_transform(self, batch_data: List[Dict]) -> Dict:
        model_inputs = {}
        # if not self.is_training:
        raw_data_dict = [d.pop('raw_data_dict') for d in batch_data]
        model_inputs['raw_data_dict'] = raw_data_dict

        batch_pixel_values = [d.pop('pixel_values') for d in batch_data if 'pixel_values' in d]
        batch_image_grid_thw = [d.pop('image_grid_thw') for d in batch_data if 'image_grid_thw' in d]
        if len(batch_pixel_values) == 0:
            vision_placeholder = self.get_vision_placeholder()
            batch_pixel_values = [vision_placeholder.get('pixel_values')]
            batch_image_grid_thw = [vision_placeholder.get('image_grid_thw')] if 'image_grid_thw' in vision_placeholder else []

        model_inputs['pixel_values'] = torch.cat(batch_pixel_values, dim=0)
        if len(batch_image_grid_thw) > 0:
            model_inputs['image_grid_thw'] = torch.cat(batch_image_grid_thw, dim=0)
    
        batch_num_images = [d.pop('num_images') for d in batch_data]
        model_inputs['num_images'] = torch.tensor(batch_num_images)
        model_inputs.update(self.pad_collator(batch_data))
        return model_inputs
    
    def __call__(self, batch_data: Union[Dict, List[Dict]]) -> Dict:
        if isinstance(batch_data, dict):
            batch_data = [batch_data]
        batch = [self.transform(d)[0] for d in batch_data]
        return self.batch_transform(batch)
    
    def get_vision_placeholder(self):
        messages = [{"role": "user", "content": [{"type": "image", "image": Image.new(mode='RGB', size=(336, 336))}]}]
        image_processing_config = self.parse_image_processing_config({})
        return self.processor(messages, image_processing_config)
    
    def get_text_placeholder(self):
        messages = [
            {"role": "user", "content": [{"type": "text", "text": "Hello!"}]},
            {"role": "assistant", "content": [{"type": "text", "text": "Thank you very much"}]},
        ]
        image_processing_config = self.parse_image_processing_config({})
        return self.processor(messages, image_processing_config)

def init_processor(processor: Union[TarsierProcessor, str]=None, config: Dict=None):
    config = DictToObject(config) if isinstance(config, dict) else config
    if isinstance(processor, str):
        sub_processor = TarsierProcessor.from_pretrained(
            processor, 
            padding_side='left',
            trust_remote_code=True
        )
    else:
        sub_processor = processor
    processor = TarsierDataProcessor(
        processor=sub_processor,
        n_frames=config.n_frames,
        max_n_frames=config.max_n_frames,
        max_pixels=config.max_pixels,
        min_pixels=config.min_pixels,
        max_seq_len=config.max_seq_len,
        is_training=config.is_training,
        print_data_error=config.print_data_error,
        do_image_padding=config.do_image_padding,
        do_image_crop=config.do_image_crop,
        do_image_resize=config.do_image_resize,
        video_sampling_strategy=config.video_sampling_strategy,
        prompt=config.prompt,
        train_task=config.train_task
    )
    return processor

class TarsierDataset(Dataset):
    def __init__(self, ann_path="", anns=None, config: Dict=None, processor: Union[TarsierDataProcessor, TarsierProcessor, str]=None):
        self.config = DictToObject(config) if isinstance(config, dict) else config
        if not isinstance(processor, TarsierDataProcessor):
            self.processor = init_processor(processor, config)
        else:
            self.processor = processor
        if anns is None:
            self.anns = []
            if isinstance(ann_path, str):
                ann_path = [ann_path]
            for path in ann_path:
                self.anns.extend(read_jsonlines(path))
        else:
            self.anns = anns

    def __len__(self):
        return len(self.anns)
    
    def __getitem__(self, index):
        if index < 0 or index >= len(self.anns):
            raise IndexError("Index out of range")
        try:
            ann = self.anns[index]
            model_inputs = self.processor(ann)
        except Exception as e:
            print(f"Load data error: {e}")
            return ann, None
        return ann, model_inputs



"""D1/D2 Point Cloud PSNR (MPEG Standard). pc_d1_psnr, pc_d2_psnr — dB, higher = better"""
import logging, numpy as np
from pathlib import Path
from typing import Optional
from ayase.models import Sample, QualityMetrics
from ayase.base_modules import ReferenceBasedModule
logger = logging.getLogger(__name__)

class PCPSNRModule(ReferenceBasedModule):
    name = "pc_psnr"; description = "D1/D2 MPEG point cloud PSNR"; metric_field = None; default_config = {}
    def process(self, sample):
        ref = getattr(sample, "reference_path", None)
        if ref is None: return sample
        if not isinstance(ref, Path): ref = Path(ref)
        if not ref.exists(): return sample
        ext = sample.path.suffix.lower()
        if ext not in (".ply", ".pcd"): return sample
        try:
            d1, d2 = self._compute(sample.path, ref)
            if d1 is not None:
                if sample.quality_metrics is None: sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.pc_d1_psnr = d1
                sample.quality_metrics.pc_d2_psnr = d2
        except Exception as e:
            logger.warning(f"PC-PSNR failed: {e}")
        return sample
    def compute_reference_score(self, sample_path, reference_path):
        d1, d2 = self._compute(sample_path, reference_path)
        return d1
    def _compute(self, sample_path, ref_path):
        try:
            import open3d as o3d
            pc1 = o3d.io.read_point_cloud(str(sample_path))
            pc2 = o3d.io.read_point_cloud(str(ref_path))
            p1 = np.asarray(pc1.points); p2 = np.asarray(pc2.points)
            if len(p1) == 0 or len(p2) == 0: return None, None
            # D1: point-to-point
            from scipy.spatial import cKDTree
            tree2 = cKDTree(p2); d_p2p, _ = tree2.query(p1)
            mse_d1 = np.mean(d_p2p**2)
            peak = np.max(np.linalg.norm(p2.max(axis=0)-p2.min(axis=0)))
            d1 = 10*np.log10(peak**2/max(mse_d1, 1e-10))
            # D2: point-to-plane (approximation using normals)
            d2 = None  # Requires normals for point-to-plane
            if pc2.has_normals():
                normals = np.asarray(pc2.normals)
                _, idx = tree2.query(p1)
                proj = np.sum((p1 - p2[idx]) * normals[idx], axis=1)**2
                mse_d2 = np.mean(proj)
                d2 = 10*np.log10(peak**2/max(mse_d2, 1e-10))
            return float(d1), float(d2)
        except ImportError:
            logger.debug("open3d/scipy not installed for PC-PSNR")
            return None, None



"""Harmful Content Detection module.

Beyond NSFW — detects potentially harmful visual content:

  harmful_content_score — 0-1 (higher = more harmful)

Detection categories:
  - Violence / fighting (motion + red-channel analysis)
  - Blood / gore (colour-based heuristic)
  - Disturbing textures (high-frequency noise patterns)

Uses OpenCV heuristics as baseline.  If ``transformers`` and a
safety classifier are available, uses zero-shot CLIP classification
for more accurate results.
"""

import logging
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class HarmfulContentModule(PipelineModule):
    name = "harmful_content"
    description = "Violence, gore, and disturbing content detection"
    default_config = {
        "subsample": 10,
        "max_frames": 60,
        "warning_threshold": 0.4,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 10)
        self.max_frames = self.config.get("max_frames", 60)
        self.warning_threshold = self.config.get("warning_threshold", 0.4)

        self._clip_model = None
        self._clip_processor = None
        self._clip_available = False

    def setup(self) -> None:
        try:
            import torch
            from transformers import CLIPModel, CLIPProcessor

            device = "cuda" if torch.cuda.is_available() else "cpu"
            self._clip_model = CLIPModel.from_pretrained(
                "openai/clip-vit-base-patch32"
            ).to(device).eval()
            self._clip_processor = CLIPProcessor.from_pretrained(
                "openai/clip-vit-base-patch32"
            )
            self._clip_available = True
            logger.info(f"Harmful content: CLIP classifier on {device}")
        except ImportError:
            logger.info("CLIP unavailable, using heuristic-only detection")
        except Exception as e:
            logger.warning(f"CLIP init failed: {e}")

    # ------------------------------------------------------------------
    # Heuristic detectors
    # ------------------------------------------------------------------

    @staticmethod
    def _blood_gore_score(frame_bgr: np.ndarray) -> float:
        """Heuristic: detect red-dominant regions (blood/gore-like).

        Looks for high red saturation in HSV space.
        Returns 0-1.
        """
        hsv = cv2.cvtColor(frame_bgr, cv2.COLOR_BGR2HSV)

        # Red hue ranges (wraps around 0/180 in OpenCV HSV)
        mask_low = cv2.inRange(hsv, (0, 100, 50), (10, 255, 255))
        mask_high = cv2.inRange(hsv, (170, 100, 50), (180, 255, 255))
        red_mask = mask_low | mask_high

        red_ratio = float(red_mask.sum() / 255 / red_mask.size)

        # Dark red (blood-like) specifically
        dark_red = cv2.inRange(hsv, (0, 100, 30), (15, 255, 120))
        dark_red_high = cv2.inRange(hsv, (165, 100, 30), (180, 255, 120))
        blood_mask = dark_red | dark_red_high
        blood_ratio = float(blood_mask.sum() / 255 / blood_mask.size)

        # Combined: heavy red presence + dark red = higher score
        combined = red_ratio * 0.4 + blood_ratio * 0.6
        return float(np.clip(combined * 5.0, 0, 1))

    @staticmethod
    def _violence_motion_score(
        prev_gray: np.ndarray, curr_gray: np.ndarray
    ) -> float:
        """Heuristic: high-magnitude sudden motion (violence indicator).

        Rapid motion alone doesn't indicate violence, but combined
        with other signals it contributes.
        Returns 0-1.
        """
        diff = cv2.absdiff(prev_gray, curr_gray)
        motion_intensity = float(diff.mean()) / 255.0
        # Very high motion (>0.15) is suspicious
        return float(np.clip(motion_intensity * 5.0, 0, 1))

    # ------------------------------------------------------------------
    # CLIP zero-shot classification
    # ------------------------------------------------------------------

    def _clip_harm_score(self, frame_bgr: np.ndarray) -> Optional[float]:
        if not self._clip_available:
            return None

        try:
            import torch
            from PIL import Image

            rgb = cv2.cvtColor(frame_bgr, cv2.COLOR_BGR2RGB)
            pil_img = Image.fromarray(rgb)

            safe_texts = [
                "a safe, normal photograph",
                "a peaceful scene",
                "a landscape or everyday scene",
            ]
            harmful_texts = [
                "a violent scene with fighting",
                "a disturbing or graphic image",
                "blood or injury",
            ]

            all_texts = safe_texts + harmful_texts
            inputs = self._clip_processor(
                text=all_texts, images=pil_img, return_tensors="pt", padding=True
            )
            device = next(self._clip_model.parameters()).device
            inputs = {k: v.to(device) for k, v in inputs.items()}

            with torch.no_grad():
                outputs = self._clip_model(**inputs)
                probs = torch.softmax(outputs.logits_per_image[0], dim=0).cpu().numpy()

            # Sum of harmful probabilities
            n_safe = len(safe_texts)
            harm_prob = float(probs[n_safe:].sum())
            return harm_prob

        except Exception as e:
            logger.debug(f"CLIP harm classification failed: {e}")
            return None

    # ------------------------------------------------------------------
    # Process
    # ------------------------------------------------------------------

    def _score_frame(
        self,
        frame_bgr: np.ndarray,
        prev_gray: Optional[np.ndarray] = None,
    ) -> float:
        blood = self._blood_gore_score(frame_bgr)

        motion = 0.0
        if prev_gray is not None:
            curr_gray = cv2.cvtColor(frame_bgr, cv2.COLOR_BGR2GRAY)
            motion = self._violence_motion_score(prev_gray, curr_gray)

        clip_score = self._clip_harm_score(frame_bgr)

        if clip_score is not None:
            return 0.2 * blood + 0.1 * motion + 0.7 * clip_score
        return 0.6 * blood + 0.4 * motion

    def process(self, sample: Sample) -> Sample:
        try:
            if sample.is_video:
                score = self._process_video(sample.path)
            else:
                score = self._process_image(sample.path)

            if score is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.harmful_content_score = score

            if score > self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Potentially harmful content detected: {score:.2f}",
                        details={"harmful_content_score": score},
                        recommendation=(
                            "Content may contain violence, gore, or "
                            "disturbing material.  Review manually."
                        ),
                    )
                )

            logger.debug(f"Harmful content for {sample.path.name}: {score:.3f}")

        except Exception as e:
            logger.error(f"Harmful content detection failed for {sample.path}: {e}")

        return sample

    def _process_image(self, path: Path) -> Optional[float]:
        img = cv2.imread(str(path))
        if img is None:
            return None
        return self._score_frame(img)

    def _process_video(self, path: Path) -> Optional[float]:
        cap = cv2.VideoCapture(str(path))
        if not cap.isOpened():
            return None

        scores = []
        prev_gray = None
        idx = 0

        while idx < self.max_frames:
            ret, frame = cap.read()
            if not ret:
                break
            if idx % self.subsample == 0:
                s = self._score_frame(frame, prev_gray)
                scores.append(s)
                prev_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            idx += 1

        cap.release()
        return float(np.mean(scores)) if scores else None



"""Tests for camera_jitter module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_camera_jitter_basics():
    from ayase.modules.camera_jitter import CameraJitterModule
    _test_module_basics(CameraJitterModule, "camera_jitter")

def test_camera_jitter_video(video_sample):
    from ayase.modules.camera_jitter import CameraJitterModule
    video_sample.quality_metrics = QualityMetrics()
    m = CameraJitterModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""SFID — Spatial FID (Fréchet Inception Distance).

Dataset-level distribution metric via pyiqa. Computes FID on
spatially-aware features rather than global features, capturing
local texture and structure quality differences.

sfid_score — LOWER = better (closer distributions)
"""

import logging
from typing import Optional, List

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import BatchMetricModule

logger = logging.getLogger(__name__)


class SFIDModule(BatchMetricModule):
    name = "sfid"
    description = "SFID spatial Fréchet Inception Distance via pyiqa (batch, lower=better)"
    default_config = {
        "spatial_patches": 4,
        "feature_dim": 128,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.spatial_patches = self.config.get("spatial_patches", 4)
        self.feature_dim = self.config.get("feature_dim", 128)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try pyiqa SFID
        try:
            import pyiqa
            self._model = pyiqa.create_metric("sfid", device="cpu")
            self._backend = "pyiqa"
            logger.info("SFID (pyiqa) initialised")
            return
        except (ImportError, Exception):
            pass

        # Tier 2: Heuristic fallback
        self._backend = "heuristic"
        logger.info("SFID (heuristic) initialised — install pyiqa for full model")

    def extract_features(self, sample: Sample) -> Optional[np.ndarray]:
        """Extract spatial patch features from a sample."""
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return None
                mid = total // 2
                cap.set(cv2.CAP_PROP_POS_FRAMES, mid)
                ret, frame = cap.read()
                if not ret:
                    return None
            finally:
                cap.release()
        else:
            frame = cv2.imread(str(sample.path))
            if frame is None:
                return None

        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
        h, w = gray.shape

        # Spatial patch features
        n = self.spatial_patches
        patch_h = h // n
        patch_w = w // n
        features = []

        for i in range(n):
            for j in range(n):
                patch = gray[
                    i * patch_h : (i + 1) * patch_h,
                    j * patch_w : (j + 1) * patch_w,
                ]
                if patch.size == 0:
                    features.extend([0.0] * 8)
                    continue

                # Per-patch statistics
                features.append(np.mean(patch))
                features.append(np.std(patch))

                # Gradient features
                gx = cv2.Sobel(patch, cv2.CV_64F, 1, 0, ksize=3)
                gy = cv2.Sobel(patch, cv2.CV_64F, 0, 1, ksize=3)
                grad_mag = np.sqrt(gx ** 2 + gy ** 2)
                features.append(np.mean(grad_mag))
                features.append(np.std(grad_mag))

                # Texture features
                lap = cv2.Laplacian(patch, cv2.CV_64F)
                features.append(np.mean(np.abs(lap)))
                features.append(np.var(lap))

                # Histogram features (2 bins for compactness)
                hist = np.histogram(patch, bins=2, range=(0, 256))[0]
                hist = hist.astype(np.float64) / (hist.sum() + 1e-7)
                features.extend(hist.tolist())

        features = np.array(features, dtype=np.float64)

        # Pad or truncate to fixed dimension
        if len(features) < self.feature_dim:
            features = np.pad(features, (0, self.feature_dim - len(features)))
        else:
            features = features[: self.feature_dim]

        return features

    def compute_distribution_metric(
        self, features: List, reference_features: Optional[List] = None
    ) -> float:
        """Compute spatial FID between feature distributions."""
        feat_matrix = np.array(features)

        if reference_features is not None and len(reference_features) >= 2:
            ref_matrix = np.array(reference_features)
        else:
            # Self-comparison: split into two halves
            n = len(feat_matrix)
            if n < 4:
                return 0.0
            mid = n // 2
            ref_matrix = feat_matrix[mid:]
            feat_matrix = feat_matrix[:mid]

        # Compute Fréchet distance
        mu1 = np.mean(feat_matrix, axis=0)
        mu2 = np.mean(ref_matrix, axis=0)

        sigma1 = np.cov(feat_matrix, rowvar=False) if feat_matrix.shape[0] > 1 else np.eye(feat_matrix.shape[1])
        sigma2 = np.cov(ref_matrix, rowvar=False) if ref_matrix.shape[0] > 1 else np.eye(ref_matrix.shape[1])

        # Ensure 2D
        if sigma1.ndim < 2:
            sigma1 = np.atleast_2d(sigma1)
        if sigma2.ndim < 2:
            sigma2 = np.atleast_2d(sigma2)

        # FID = ||mu1 - mu2||^2 + Tr(sigma1 + sigma2 - 2*sqrt(sigma1 @ sigma2))
        diff = mu1 - mu2
        mean_diff_sq = np.sum(diff ** 2)

        # Simplified: use trace approximation
        trace_sum = np.trace(sigma1) + np.trace(sigma2)

        # Approximate sqrt(sigma1 @ sigma2) trace
        product = sigma1 @ sigma2
        eigenvalues = np.real(np.linalg.eigvals(product))
        eigenvalues = np.maximum(eigenvalues, 0)
        trace_sqrt = np.sum(np.sqrt(eigenvalues))

        fid = mean_diff_sq + trace_sum - 2 * trace_sqrt
        return float(np.clip(fid, 0.0, 1e6))
